import json
from pathlib import Path
from typing import Literal
import copy

from .utils import is_settings, clean_value


class debug:
    def __init__(self, instance, func, mode: str = "", **kwargs):
        if not is_settings(type(instance)):
            return
        self.mode = mode
        if func == "__init__":
            self.show_init_info(instance, **kwargs)

    def show_init_info(self, instance, **kwargs):
        values = kwargs.get("values")
        if self.mode == "after":
            if instance._arfi_debug and instance.read_config:
                print("\033[34m")
                print("values after handler:")
                clear_value = clean_value(copy.deepcopy(values))
                print(json.dumps(clear_value, indent=4))
                print("\033[0m")
                return
        else:
            if instance._arfi_debug and instance.read_config or kwargs.get("arfi_debug"):
                conf_path = self._convert_debug_path(instance, instance.conf_path, "conf")
                env_path = self._convert_debug_path(instance, instance.env_path, "env")

                print("\033[34m")
                print(f"[PRE INIT] {instance.__class__.__name__}")
                print(f"root_dir = {instance.root_dir}")
                print(f"BASE_DIR = {instance.BASE_DIR}")
                print(f"pyproject_toml_path = {instance.pyproject_toml_path}")
                print("conf_path =", json.dumps(conf_path, indent=4))
                print("env_path =", json.dumps(env_path, indent=4))

                if instance._arfi_debug and instance.read_config:
                    print(f"mode_dir_path = {instance.mode_dir_path}")
                    print(f"computed_mode_dir = {instance.computed_mode_dir}")
                    print(f"source_mode_dir = {instance.source_mode_dir}")
                    print(f"nested_mode_dir = {instance.nested_mode_dir}")
                    print(f"parent_mode_dir = {instance.parent_mode_dir}")
                    print("settings_config =", instance.settings_config.model_dump_json(indent=4))
                    clear_value = clean_value(copy.deepcopy(values))
                    print("values before handler:")
                    print(json.dumps(clear_value, indent=4))

                print("\033[0m")
                return

    @staticmethod
    def _convert_debug_path(instance, list_path: list[Path], mode: Literal["conf", "env"]) -> list[str]:
        result = []
        base_dir = instance.BASE_DIR
        root_dir = instance.root_dir
        if instance.BASE_DIR is not None:
            base_dir = Path(instance.BASE_DIR).resolve()
        if instance.root_dir is not None:
            root_dir = Path(instance.root_dir).resolve()
        for path in list_path:
            if mode == "conf":
                if base_dir is not None:
                    path = base_dir / path
            elif mode == "env":
                if root_dir is not None:
                    if Path(root_dir / path).is_file():
                        path = root_dir / path
                    else:
                        if base_dir is not None:
                            path = base_dir / path
                else:
                    if base_dir is not None:
                        path = base_dir / path
            result.append(path.resolve().as_posix())
        return result



from .errors import ArFiSettingsError
from .handlers import (
    ArFiBaseHandler,
    ArFiHandler,
)
from .init_config import init_settings
from .main import ArFiSettings
from .readers import (
    ArFiBaseReader,
    ArFiReader,
)
from .types import (
    EnvConfigDict,
    FileConfigDict,
    SettingsConfigDict,
)
from .version import VERSION

__all__ = (
    "ArFiSettings",
    "ArFiReader",
    "ArFiBaseReader",
    "ArFiHandler",
    "ArFiBaseHandler",
    "ArFiSettingsError",
    "SettingsConfigDict",
    "FileConfigDict",
    "EnvConfigDict",
    "init_settings",
    "__version__",
)

__version__ = VERSION



import pydantic
import pytest

from arfi_settings import ArFiSettings, EnvConfigDict, SettingsConfigDict


# @pytest.mark.current
@pytest.mark.env_config
def test_class_env_exclude_inherit_parent_valid_input(
    cwd_to_tmp,
    env_include_exclude_expected,
    env_include_exclude_msg,
):
    class AppConfig(ArFiSettings):
        pass

    config = AppConfig()
    assert config.settings_config.env_exclude_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == []

    with pytest.raises(pydantic.ValidationError) as excinfo:

        class A(ArFiSettings):
            env_config = EnvConfigDict(
                env_exclude_inherit_parent=True,
            )

        A()
    assert excinfo.value.errors(include_url=False) == [
        {
            "type": "list_type",
            "loc": (
                "env_config",
                "env_exclude_inherit_parent",
            ),
            "msg": "Input should be a valid list",
            "input": True,
        },
    ]

    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_exclude_inherit_parent=[],
        )

    config = AppConfig()
    assert config.settings_config.env_exclude_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == []

    with pytest.raises(pydantic.ValidationError) as excinfo:

        class B(ArFiSettings):
            env_config = EnvConfigDict(
                env_exclude_inherit_parent=["test"],
            )

        B()

    assert excinfo.value.errors(include_url=False) == [
        {
            "type": "literal_error",
            "loc": (
                "env_exclude_inherit_parent",
                0,
            ),
            "ctx": {
                "expected": env_include_exclude_expected,
            },
            "msg": env_include_exclude_msg,
            "input": "test",
        },
    ]

    with pytest.raises(pydantic.ValidationError) as excinfo:

        class C(ArFiSettings):
            model_config = SettingsConfigDict(
                env_exclude_inherit_parent=True,
            )

        C()
    assert excinfo.value.errors(include_url=False) == [
        {
            "type": "list_type",
            "loc": ("env_exclude_inherit_parent",),
            "msg": "Input should be a valid list",
            "input": True,
        },
    ]

    with pytest.raises(pydantic.ValidationError) as excinfo:

        class D(ArFiSettings):
            model_config = SettingsConfigDict(
                exclude_inherit_parent=True,
            )

        D()
    assert excinfo.value.errors(include_url=False) == [
        {
            "type": "list_type",
            "loc": ("exclude_inherit_parent",),
            "msg": "Input should be a valid list",
            "input": True,
        },
    ]


# @pytest.mark.current
@pytest.mark.env_config
def test_class_env_exclude_inherit_parent(cwd_to_tmp):
    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_exclude_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            env_exclude_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_exclude_inherit_parent=["env_file"],
        )
        model_config = SettingsConfigDict(
            env_exclude_inherit_parent=["env_prefix_as_source_mode_dir"],
        )

    config = AppConfig()
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            exclude_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            env_exclude_inherit_parent=["env_file"],
            exclude_inherit_parent=["env_prefix_as_source_mode_dir"],
        )

    config = AppConfig()
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_exclude_inherit_parent=["env_file"],
        )
        model_config = SettingsConfigDict(
            env_exclude_inherit_parent=["env_nested_delimiter"],
            exclude_inherit_parent=["env_prefix_as_source_mode_dir"],
        )

    config = AppConfig()
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]


# @pytest.mark.current
@pytest.mark.env_config
def test_only_init_env_exclude_inherit_parent(
    cwd_to_tmp,
    env_include_exclude_expected,
    env_include_exclude_msg,
):
    class AppConfig(ArFiSettings):
        pass

    config = AppConfig(
        _env_exclude_inherit_parent=[],
    )
    assert config.settings_config.env_exclude_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == []

    config = AppConfig(
        _exclude_inherit_parent=[],
    )
    assert config.settings_config.env_exclude_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == []

    with pytest.raises(pydantic.ValidationError) as excinfo:
        AppConfig(
            _env_exclude_inherit_parent=True,
        )
    assert excinfo.value.errors(include_url=False) == [
        {
            "type": "list_type",
            "loc": ("_env_exclude_inherit_parent",),
            "msg": "Input should be a valid list",
            "input": True,
        },
    ]

    with pytest.raises(pydantic.ValidationError) as excinfo:
        AppConfig(
            _env_exclude_inherit_parent=["test"],
        )
    assert excinfo.value.errors(include_url=False) == [
        {
            "type": "literal_error",
            "loc": (
                "_env_exclude_inherit_parent",
                0,
            ),
            "ctx": {
                "expected": env_include_exclude_expected,
            },
            "msg": env_include_exclude_msg,
            "input": "test",
        },
    ]

    config = AppConfig(
        _env_exclude_inherit_parent=["env_file"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    config = AppConfig(
        _exclude_inherit_parent=["env_file"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    config = AppConfig(
        _env_exclude_inherit_parent=["env_file"],
        _exclude_inherit_parent=["env_prefix"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]


# @pytest.mark.current
@pytest.mark.env_config
def test_init_and_class_env_exclude_inherit_parent(cwd_to_tmp):
    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_exclude_inherit_parent=["env_file"],
        )

    config = AppConfig(
        _env_exclude_inherit_parent=["env_prefix_as_source_mode_dir"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]

    config = AppConfig(
        _exclude_inherit_parent=["env_prefix_as_source_mode_dir"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]

    config = AppConfig(
        _env_exclude_inherit_parent=["env_prefix_as_source_mode_dir"],
        _exclude_inherit_parent=["env_prefix"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            env_exclude_inherit_parent=["env_file"],
        )

    config = AppConfig(
        _env_exclude_inherit_parent=["env_prefix_as_source_mode_dir"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]

    config = AppConfig(
        _exclude_inherit_parent=["env_prefix_as_source_mode_dir"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            env_exclude_inherit_parent=["env_file"],
        )

    config = AppConfig(
        _env_exclude_inherit_parent=["env_prefix_as_source_mode_dir"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]

    config = AppConfig(
        _exclude_inherit_parent=["env_prefix_as_source_mode_dir"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]

    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_exclude_inherit_parent=["env_file"],
        )
        model_config = SettingsConfigDict(
            env_exclude_inherit_parent=["env_prefix_as_source_mode_dir"],
        )

    config = AppConfig(
        _env_exclude_inherit_parent=["env_prefix"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_prefix"]
    assert config.settings_config.exclude_inherit_parent == ["env_prefix"]

    config = AppConfig(
        _exclude_inherit_parent=["env_prefix"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_prefix"]
    assert config.settings_config.exclude_inherit_parent == ["env_prefix"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            exclude_inherit_parent=["env_file"],
        )

    config = AppConfig(
        _env_exclude_inherit_parent=["env_prefix_as_source_mode_dir"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]

    config = AppConfig(
        _exclude_inherit_parent=["env_prefix_as_source_mode_dir"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.exclude_inherit_parent == ["env_prefix_as_source_mode_dir"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            env_exclude_inherit_parent=["env_file"],
            exclude_inherit_parent=["env_prefix_as_source_mode_dir"],
        )

    config = AppConfig(
        _env_exclude_inherit_parent=["env_prefix"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_prefix"]
    assert config.settings_config.exclude_inherit_parent == ["env_prefix"]

    config = AppConfig(
        _exclude_inherit_parent=["env_prefix"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_prefix"]
    assert config.settings_config.exclude_inherit_parent == ["env_prefix"]

    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_exclude_inherit_parent=["env_file"],
        )
        model_config = SettingsConfigDict(
            env_exclude_inherit_parent=["env_nested_delimiter"],
            exclude_inherit_parent=["env_prefix_as_source_mode_dir"],
        )

    config = AppConfig(
        _env_exclude_inherit_parent=["env_prefix"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_prefix"]
    assert config.settings_config.exclude_inherit_parent == ["env_prefix"]

    config = AppConfig(
        _exclude_inherit_parent=["env_prefix"],
    )
    assert config.settings_config.env_exclude_inherit_parent == ["env_prefix"]
    assert config.settings_config.exclude_inherit_parent == ["env_prefix"]


# @pytest.mark.current
@pytest.mark.env_config
def test_class_env_include_inherit_parent_valid_input(
    cwd_to_tmp,
    env_include_exclude_expected,
    env_include_exclude_msg,
):
    class AppConfig(ArFiSettings):
        pass

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.include_inherit_parent == []

    with pytest.raises(pydantic.ValidationError) as excinfo:

        class A(ArFiSettings):
            env_config = EnvConfigDict(
                env_include_inherit_parent=True,
            )

        A()

    assert excinfo.value.errors(include_url=False) == [
        {
            "type": "list_type",
            "loc": (
                "env_config",
                "env_include_inherit_parent",
            ),
            "msg": "Input should be a valid list",
            "input": True,
        },
    ]

    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_include_inherit_parent=[],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.include_inherit_parent == []

    with pytest.raises(pydantic.ValidationError) as excinfo:

        class B(ArFiSettings):
            env_config = EnvConfigDict(
                env_include_inherit_parent=["test"],
            )

        B()
    assert excinfo.value.errors(include_url=False) == [
        {
            "type": "literal_error",
            "loc": (
                "env_include_inherit_parent",
                0,
            ),
            "ctx": {
                "expected": env_include_exclude_expected,
            },
            "msg": env_include_exclude_msg,
            "input": "test",
        },
    ]


# @pytest.mark.current
@pytest.mark.env_config
def test_class_env_include_inherit_parent(cwd_to_tmp):
    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_include_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            env_include_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_include_inherit_parent=["env_file"],
        )
        model_config = SettingsConfigDict(
            env_include_inherit_parent=["env_prefix_as_source_mode_dir"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            env_include_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            env_include_inherit_parent=["env_file"],
            include_inherit_parent=["env_prefix_as_source_mode_dir"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_include_inherit_parent=["env_file"],
        )
        model_config = SettingsConfigDict(
            env_include_inherit_parent=["env_nested_delimiter"],
            include_inherit_parent=["env_prefix_as_source_mode_dir"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_file"]


# @pytest.mark.current
@pytest.mark.env_config
def test_only_init_env_include_inherit_parent(
    cwd_to_tmp,
    env_include_exclude_expected,
    env_include_exclude_msg,
):
    class AppConfig(ArFiSettings):
        pass

    config = AppConfig(
        _env_include_inherit_parent=[],
    )
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.include_inherit_parent == []

    config = AppConfig(
        _include_inherit_parent=[],
    )
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.include_inherit_parent == []

    with pytest.raises(pydantic.ValidationError) as excinfo:
        AppConfig(
            _env_include_inherit_parent=True,
        )
    assert excinfo.value.errors(include_url=False) == [
        {
            "type": "list_type",
            "loc": ("_env_include_inherit_parent",),
            "msg": "Input should be a valid list",
            "input": True,
        },
    ]

    with pytest.raises(pydantic.ValidationError) as excinfo:
        AppConfig(
            _env_include_inherit_parent=["test"],
        )
    assert excinfo.value.errors(include_url=False) == [
        {
            "type": "literal_error",
            "loc": (
                "_env_include_inherit_parent",
                0,
            ),
            "ctx": {
                "expected": env_include_exclude_expected,
            },
            "msg": env_include_exclude_msg,
            "input": "test",
        },
    ]

    config = AppConfig(
        _env_include_inherit_parent=["env_file"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_file"]

    config = AppConfig(
        _env_include_inherit_parent=["env_file"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_file"]

    config = AppConfig(
        _env_include_inherit_parent=["env_file"],
        _include_inherit_parent=["env_prefix"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_file"]


# @pytest.mark.current
@pytest.mark.env_config
def test_init_and_class_env_include_inherit_parent(cwd_to_tmp):
    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_include_inherit_parent=["env_file"],
        )

    config = AppConfig(
        _env_include_inherit_parent=["env_prefix_as_source_mode_dir"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.include_inherit_parent == ["env_prefix_as_source_mode_dir"]

    config = AppConfig(
        _include_inherit_parent=["env_prefix_as_source_mode_dir"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.include_inherit_parent == ["env_prefix_as_source_mode_dir"]

    config = AppConfig(
        _env_include_inherit_parent=["env_prefix_as_source_mode_dir"],
        _include_inherit_parent=["env_prefix"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.include_inherit_parent == ["env_prefix_as_source_mode_dir"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            env_include_inherit_parent=["env_file"],
        )

    config = AppConfig(
        _env_include_inherit_parent=["env_prefix_as_source_mode_dir"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.include_inherit_parent == ["env_prefix_as_source_mode_dir"]

    config = AppConfig(
        _include_inherit_parent=["env_prefix_as_source_mode_dir"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.include_inherit_parent == ["env_prefix_as_source_mode_dir"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            env_include_inherit_parent=["env_file"],
        )

    config = AppConfig(
        _env_include_inherit_parent=["env_prefix_as_source_mode_dir"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.include_inherit_parent == ["env_prefix_as_source_mode_dir"]

    config = AppConfig(
        _include_inherit_parent=["env_prefix_as_source_mode_dir"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.include_inherit_parent == ["env_prefix_as_source_mode_dir"]

    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_include_inherit_parent=["env_file"],
        )
        model_config = SettingsConfigDict(
            env_include_inherit_parent=["env_prefix_as_source_mode_dir"],
        )

    config = AppConfig(
        _env_include_inherit_parent=["env_prefix"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_prefix"]
    assert config.settings_config.include_inherit_parent == ["env_prefix"]

    config = AppConfig(
        _include_inherit_parent=["env_prefix"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_prefix"]
    assert config.settings_config.include_inherit_parent == ["env_prefix"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            include_inherit_parent=["env_file"],
        )

    config = AppConfig(
        _env_include_inherit_parent=["env_prefix_as_source_mode_dir"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.include_inherit_parent == ["env_prefix_as_source_mode_dir"]

    config = AppConfig(
        _include_inherit_parent=["env_prefix_as_source_mode_dir"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_prefix_as_source_mode_dir"]
    assert config.settings_config.include_inherit_parent == ["env_prefix_as_source_mode_dir"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            env_include_inherit_parent=["env_file"],
            include_inherit_parent=["env_prefix_as_source_mode_dir"],
        )

    config = AppConfig(
        _env_include_inherit_parent=["env_prefix"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_prefix"]
    assert config.settings_config.include_inherit_parent == ["env_prefix"]

    config = AppConfig(
        _include_inherit_parent=["env_prefix"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_prefix"]
    assert config.settings_config.include_inherit_parent == ["env_prefix"]

    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_include_inherit_parent=["env_file"],
        )
        model_config = SettingsConfigDict(
            env_include_inherit_parent=["env_nested_delimiter"],
            include_inherit_parent=["env_prefix_as_source_mode_dir"],
        )

    config = AppConfig(
        _env_include_inherit_parent=["env_prefix"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_prefix"]
    assert config.settings_config.include_inherit_parent == ["env_prefix"]

    config = AppConfig(
        _include_inherit_parent=["env_prefix"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_prefix"]
    assert config.settings_config.include_inherit_parent == ["env_prefix"]


# @pytest.mark.current
@pytest.mark.env_config
def test_combination_class_env_include_exclude_parent(cwd_to_tmp):
    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_include_inherit_parent=["env_file"],
            env_exclude_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_include_inherit_parent=["env_file"],
        )
        model_config = SettingsConfigDict(
            env_exclude_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_include_inherit_parent=["env_file"],
        )
        model_config = SettingsConfigDict(
            exclude_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            env_include_inherit_parent=["env_file"],
            env_exclude_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            env_include_inherit_parent=["env_file"],
            exclude_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            include_inherit_parent=["env_file"],
            exclude_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_exclude_inherit_parent=[],
        )
        model_config = SettingsConfigDict(
            include_inherit_parent=["env_file"],
            exclude_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_include_inherit_parent=["env_file", "env_ignore_missing"],
            env_exclude_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_include_inherit_parent=["env_file", "env_ignore_missing"],
        )
        model_config = SettingsConfigDict(
            env_exclude_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_include_inherit_parent=["env_file", "env_ignore_missing"],
        )
        model_config = SettingsConfigDict(
            exclude_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            env_include_inherit_parent=["env_file", "env_ignore_missing"],
            env_exclude_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            env_include_inherit_parent=["env_file", "env_ignore_missing"],
            exclude_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            include_inherit_parent=["env_file", "env_ignore_missing"],
            exclude_inherit_parent=["env_file"],
        )

    config = AppConfig()
    assert config.settings_config.env_include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]


# @pytest.mark.current
@pytest.mark.env_config
def test_combination_init_and_class_env_include_exclude_parent(cwd_to_tmp):
    # 0
    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_include_inherit_parent=["env_file"],
            env_exclude_inherit_parent=["env_file"],
        )

    config = AppConfig(
        _env_include_inherit_parent=["env_file"],
    )
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    config = AppConfig(
        _env_include_inherit_parent=["env_file"],
        _env_exclude_inherit_parent=["env_file"],
    )
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    config = AppConfig(
        _env_include_inherit_parent=["env_file"],
        _exclude_inherit_parent=["env_file"],
    )
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    config = AppConfig(
        _env_exclude_inherit_parent=[],
    )
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.env_exclude_inherit_parent == []
    assert config.settings_config.include_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == []

    config = AppConfig(
        _exclude_inherit_parent=[],
    )
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    config = AppConfig(
        _exclude_inherit_parent=["env_ignore_missing"],
    )
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.env_exclude_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.include_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == ["env_ignore_missing"]

    config = AppConfig(
        _env_include_inherit_parent=["env_ignore_missing"],
        _env_exclude_inherit_parent=[],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.env_exclude_inherit_parent == []
    assert config.settings_config.include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.exclude_inherit_parent == []

    config = AppConfig(
        _env_include_inherit_parent=["env_ignore_missing"],
        _exclude_inherit_parent=[],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    config = AppConfig(
        _env_include_inherit_parent=["env_ignore_missing"],
        _exclude_inherit_parent=["env_ignore_missing"],
    )
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.env_exclude_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.include_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == ["env_ignore_missing"]

    config = AppConfig(
        _include_inherit_parent=["env_file"],
        _exclude_inherit_parent=["env_ignore_missing"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_file"]
    assert config.settings_config.env_exclude_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.include_inherit_parent == ["env_file"]
    assert config.settings_config.exclude_inherit_parent == ["env_ignore_missing"]

    config = AppConfig(
        _include_inherit_parent=["env_file"],
        _exclude_inherit_parent=["env_ignore_missing", "env_file"],
    )
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.env_exclude_inherit_parent == ["env_ignore_missing", "env_file"]
    assert config.settings_config.include_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == ["env_ignore_missing", "env_file"]

    # 1
    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_include_inherit_parent=["env_file"],
        )
        model_config = SettingsConfigDict(
            env_exclude_inherit_parent=["env_file"],
        )

    config = AppConfig(
        _env_include_inherit_parent=["env_ignore_missing"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    config = AppConfig(
        _include_inherit_parent=["env_ignore_missing"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    config = AppConfig(
        _include_inherit_parent=["env_ignore_missing"],
        _env_exclude_inherit_parent=[],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.env_exclude_inherit_parent == []
    assert config.settings_config.include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.exclude_inherit_parent == []

    config = AppConfig(
        _include_inherit_parent=["env_ignore_missing"],
        _exclude_inherit_parent=[],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    # 2
    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_include_inherit_parent=["env_file"],
        )
        model_config = SettingsConfigDict(
            exclude_inherit_parent=["env_file"],
        )

    config = AppConfig(
        _include_inherit_parent=["env_ignore_missing"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.env_exclude_inherit_parent == ["env_file"]
    assert config.settings_config.include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.exclude_inherit_parent == ["env_file"]

    config = AppConfig(
        _env_include_inherit_parent=["env_ignore_missing"],
        _env_exclude_inherit_parent=[],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.env_exclude_inherit_parent == []
    assert config.settings_config.include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.exclude_inherit_parent == []

    # 3
    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            env_include_inherit_parent=["env_file"],
            env_exclude_inherit_parent=["env_file"],
        )

    config = AppConfig(
        _include_inherit_parent=["env_ignore_missing"],
        _env_exclude_inherit_parent=[],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.env_exclude_inherit_parent == []
    assert config.settings_config.include_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.exclude_inherit_parent == []

    # 4
    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_include_inherit_parent=["env_file", "env_ignore_missing"],
            env_exclude_inherit_parent=["env_nested_delimiter"],
        )

    config = AppConfig(
        _include_inherit_parent=["env_nested_delimiter"],
    )
    assert config.settings_config.env_include_inherit_parent == []
    assert config.settings_config.env_exclude_inherit_parent == ["env_nested_delimiter"]
    assert config.settings_config.include_inherit_parent == []
    assert config.settings_config.exclude_inherit_parent == ["env_nested_delimiter"]

    config = AppConfig(
        _exclude_inherit_parent=["env_ignore_missing"],
    )
    assert config.settings_config.env_include_inherit_parent == ["env_file"]
    assert config.settings_config.env_exclude_inherit_parent == ["env_ignore_missing"]
    assert config.settings_config.include_inherit_parent == ["env_file"]
    assert config.settings_config.exclude_inherit_parent == ["env_ignore_missing"]


# @pytest.mark.current
@pytest.mark.env_config
def test_env_include_inherit_parent(cwd_to_tmp):
    class Proxy(ArFiSettings):
        pass

    class AppSettings(ArFiSettings):
        proxy: Proxy = Proxy()

    class AppConfig(ArFiSettings):
        app: AppSettings
        model_config = SettingsConfigDict(
            env_prefix_as_mode_dir=True,
            env_file_encoding="cp1251",
        )

    config = AppConfig()
    assert config.app.settings_config.env_prefix == "app_"
    assert config.app.settings_config.env_file_encoding == "cp1251"
    assert config.app.proxy.settings_config.env_prefix == "app_proxy_"
    assert config.app.proxy.settings_config.env_file_encoding == "cp1251"

    class AppSettings(ArFiSettings):
        proxy: Proxy = Proxy()
        model_config = SettingsConfigDict(
            env_include_inherit_parent=["env_file_encoding"],
        )

    class AppConfig(ArFiSettings):
        app: AppSettings
        model_config = SettingsConfigDict(
            env_prefix_as_mode_dir=True,
            env_file_encoding="cp1251",
        )

    config = AppConfig()
    assert config.app.settings_config.env_prefix == ""
    assert config.app.settings_config.env_file_encoding == "cp1251"
    assert config.app.proxy.settings_config.env_prefix == ""
    assert config.app.proxy.settings_config.env_file_encoding == "cp1251"

    class AppSettings(ArFiSettings):
        proxy: Proxy = Proxy()
        model_config = SettingsConfigDict(
            env_include_inherit_parent=["env_include_inherit_parent"],
        )

    class AppConfig(ArFiSettings):
        app: AppSettings
        model_config = SettingsConfigDict(
            env_prefix_as_mode_dir=True,
            env_file_encoding="cp1251",
            env_nested_delimiter=".",
            include_inherit_parent=["env_file_encoding"],
        )

    config = AppConfig()
    assert config.app.settings_config.env_prefix == ""
    assert config.app.settings_config.env_file_encoding == "cp1251"
    assert config.app.settings_config.env_nested_delimiter == ""
    assert config.app.proxy.settings_config.env_prefix == ""
    assert config.app.proxy.settings_config.env_file_encoding == "cp1251"
    assert config.app.proxy.settings_config.env_nested_delimiter == ""

    class AppSettings(ArFiSettings):
        proxy: Proxy = Proxy()
        model_config = SettingsConfigDict(
            include_inherit_parent=["env_include_inherit_parent"],
        )

    class AppConfig(ArFiSettings):
        app: AppSettings
        model_config = SettingsConfigDict(
            env_prefix_as_mode_dir=True,
            env_file_encoding="cp1251",
            env_nested_delimiter=".",
            include_inherit_parent=["env_file_encoding"],
        )

    config = AppConfig()
    assert config.settings_config.env_prefix == ""
    assert config.settings_config.env_file_encoding == "cp1251"
    assert config.settings_config.env_nested_delimiter == "."
    assert config.app.settings_config.env_prefix == ""
    assert config.app.settings_config.env_file_encoding == "cp1251"
    assert config.app.settings_config.env_nested_delimiter == ""
    assert config.app.proxy.settings_config.env_prefix == ""
    assert config.app.proxy.settings_config.env_file_encoding == "cp1251"
    assert config.app.proxy.settings_config.env_nested_delimiter == ""

    config = AppConfig(_env_file_encoding="utf-8")
    assert config.settings_config.env_prefix == ""
    assert config.settings_config.env_file_encoding == "utf-8"
    assert config.settings_config.env_nested_delimiter == "."
    assert config.app.settings_config.env_prefix == ""
    assert config.app.settings_config.env_file_encoding == "utf-8"
    assert config.app.settings_config.env_nested_delimiter == ""
    assert config.app.proxy.settings_config.env_prefix == ""
    assert config.app.proxy.settings_config.env_file_encoding == "utf-8"
    assert config.app.proxy.settings_config.env_nested_delimiter == ""

    config = AppConfig(
        _env_file_encoding="utf-8",
        app=AppSettings(
            _env_file_encoding="iso-8859-7",
        ),
    )
    assert config.settings_config.env_prefix == ""
    assert config.settings_config.env_file_encoding == "utf-8"
    assert config.settings_config.env_nested_delimiter == "."
    assert config.app.settings_config.env_prefix == ""
    assert config.app.settings_config.env_file_encoding == "iso-8859-7"
    assert config.app.settings_config.env_nested_delimiter == ""
    assert config.app.proxy.settings_config.env_prefix == ""
    assert config.app.proxy.settings_config.env_file_encoding == "iso-8859-7"
    assert config.app.proxy.settings_config.env_nested_delimiter == ""


# @pytest.mark.current
@pytest.mark.skip
@pytest.mark.todo
@pytest.mark.env_config
def test_env_include_inherit_parent_todo(cwd_to_tmp):
    # TODO: Implement !!!
    class Proxy(ArFiSettings):
        file_config_inherit_parent = False
        model_config = SettingsConfigDict(
            include_inherit_parent=["env_nested_delimiter"],
            exclude_inherit_parent=[
                "env_include_inherit_parent",
                "env_exclude_inherit_parent",
                "env_file_encoding",
            ],
        )

    class AppSettings(ArFiSettings):
        proxy: Proxy = Proxy()
        model_config = SettingsConfigDict(
            include_inherit_parent=["env_include_inherit_parent"],
        )

    class AppConfig(ArFiSettings):
        app: AppSettings
        model_config = SettingsConfigDict(
            env_prefix_as_mode_dir=True,
            env_file_encoding="cp1251",
            env_nested_delimiter=".",
            include_inherit_parent=["env_file_encoding"],
        )

    config = AppConfig(
        _env_file_encoding="utf-8",
        app=AppSettings(
            _env_file_encoding="iso-8859-7",
            _env_nested_delimiter="---@---",
        ),
    )
    assert config.settings_config.env_prefix == ""
    assert config.settings_config.env_file_encoding == "utf-8"
    assert config.settings_config.env_nested_delimiter == "."
    assert config.app.settings_config.env_prefix == ""
    assert config.app.settings_config.env_file_encoding == "iso-8859-7"
    assert config.app.settings_config.env_nested_delimiter == "---@---"
    assert config.app.proxy.settings_config.env_prefix == ""
    assert config.app.proxy.settings_config.env_file_encoding is None
    assert config.app.proxy.settings_config.env_nested_delimiter == "---@---"

    class Proxy(ArFiSettings):
        file_config_inherit_parent = False

    class AppSettings(ArFiSettings):
        proxy: Proxy
        model_config = SettingsConfigDict(
            include_inherit_parent=["env_include_inherit_parent"],
        )

    class AppConfig(ArFiSettings):
        app: AppSettings
        model_config = SettingsConfigDict(
            env_prefix_as_mode_dir=True,
            env_file_encoding="cp1251",
            env_nested_delimiter=".",
            include_inherit_parent=["env_file_encoding"],
        )

    config = AppConfig(
        _env_file_encoding="utf-8",
        app=AppSettings(
            _env_file_encoding="iso-8859-7",
            _env_nested_delimiter="---@---",
            proxy=Proxy(
                _include_inherit_parent=["env_nested_delimiter"],
                _exclude_inherit_parent=[
                    "env_include_inherit_parent",
                    "env_exclude_inherit_parent",
                    "env_file_encoding",
                ],
            ),
        ),
    )
    assert config.settings_config.env_prefix == ""
    assert config.settings_config.env_file_encoding == "utf-8"
    assert config.settings_config.env_nested_delimiter == "."
    assert config.app.settings_config.env_prefix == ""
    assert config.app.settings_config.env_file_encoding == "iso-8859-7"
    assert config.app.settings_config.env_nested_delimiter == "---@---"
    assert config.app.proxy.settings_config.env_prefix == ""
    assert config.app.proxy.settings_config.env_file_encoding is None
    assert config.app.proxy.settings_config.env_nested_delimiter == "---@---"


# @pytest.mark.current
@pytest.mark.env_config
def test_env_exclude_inherit_parent(cwd_to_tmp):
    class Proxy(ArFiSettings):
        pass

    class AppSettings(ArFiSettings):
        proxy: Proxy = Proxy()

    class AppConfig(ArFiSettings):
        app: AppSettings
        model_config = SettingsConfigDict(
            env_prefix_as_mode_dir=True,
            env_file_encoding="cp1251",
        )

    config = AppConfig()
    assert config.settings_config.env_prefix == ""
    assert config.settings_config.env_file_encoding == "cp1251"
    assert config.app.settings_config.env_prefix == "app_"
    assert config.app.settings_config.env_file_encoding == "cp1251"
    assert config.app.proxy.settings_config.env_prefix == "app_proxy_"
    assert config.app.proxy.settings_config.env_file_encoding == "cp1251"


# @pytest.mark.current
@pytest.mark.env_config
def test_multiple_reverse_inheritance_env_exclude_inherit_parent(cwd_to_tmp):
    class Address(ArFiSettings):
        model_config = SettingsConfigDict(
            env_exclude_inherit_parent=[
                "env_exclude_inherit_parent",
                "env_prefix_as_source_mode_dir",
            ],
        )

    class City(ArFiSettings):
        address: Address
        model_config = SettingsConfigDict(
            env_prefix_as_mode_dir=True,
            env_exclude_inherit_parent=[
                "env_nested_delimiter",
                "env_exclude_inherit_parent",
                "env_prefix_as_mode_dir",
            ],
        )

    class AppSettings(ArFiSettings):
        city: City
        model_config = SettingsConfigDict(
            env_exclude_inherit_parent=[
                "env_file_encoding",
                "env_exclude_inherit_parent",
            ],
        )

    class AppConfig(ArFiSettings):
        app: AppSettings
        model_config = SettingsConfigDict(
            env_prefix_as_source_mode_dir=True,
            env_nested_delimiter="---@---",
            env_file_encoding="cp1251",
            env_file=[".env", ".env.prod"],
        )

    config = AppConfig()
    assert config.app.city.address.settings_config.env_prefix == "app_city_address_"
    assert config.app.city.address.settings_config.env_file_encoding is None
    assert config.app.city.address.settings_config.env_nested_delimiter == ""
    assert config.app.city.address.settings_config.env_file == [".env", ".env.prod"]
    # assert config.app.city.address.settings_config.env_exclude_inherit_parent == [
    #     "env_exclude_inherit_parent",
    #     "env_prefix_as_source_mode_dir",
    # ]
    # assert config.app.city.address.settings_config.exclude_inherit_parent == [
    #     "env_exclude_inherit_parent",
    #     "env_prefix_as_source_mode_dir",
    # ]
    assert config.app.city.settings_config.env_prefix == "city_"
    assert config.app.city.settings_config.env_file_encoding is None
    assert config.app.city.settings_config.env_nested_delimiter == ""
    assert config.app.city.settings_config.env_file == [".env", ".env.prod"]
    assert config.app.city.settings_config.env_include_inherit_parent == []
    assert config.app.city.settings_config.env_exclude_inherit_parent == [
        "env_nested_delimiter",
        "env_exclude_inherit_parent",
        "env_prefix_as_mode_dir",
    ]
    assert config.app.city.settings_config.exclude_inherit_parent == [
        "env_nested_delimiter",
        "env_exclude_inherit_parent",
        "env_prefix_as_mode_dir",
    ]
    assert config.app.settings_config.env_prefix == "app_"
    assert config.app.settings_config.env_file_encoding is None
    assert config.app.settings_config.env_nested_delimiter == "---@---"
    assert config.app.settings_config.env_file == [".env", ".env.prod"]
    assert config.settings_config.env_prefix == ""
    assert config.settings_config.env_file_encoding == "cp1251"
    assert config.settings_config.env_nested_delimiter == "---@---"
    assert config.settings_config.env_file == [".env", ".env.prod"]


# @pytest.mark.current
@pytest.mark.env_config
def test_multiple_reverse_inheritance_env_include_inherit_parent(cwd_to_tmp):
    class Address(ArFiSettings):
        model_config = SettingsConfigDict(
            env_exclude_inherit_parent=[
                "env_exclude_inherit_parent",
                "env_include_inherit_parent",
            ],
            env_include_inherit_parent=["env_prefix_as_mode_dir", "env_file"],
        )

    class City(ArFiSettings):
        address: Address
        model_config = SettingsConfigDict(
            env_prefix_as_mode_dir=True,
            env_exclude_inherit_parent=[
                "env_nested_delimiter",
                "env_exclude_inherit_parent",
                "env_prefix_as_mode_dir",
            ],
        )

    class AppSettings(ArFiSettings):
        city: City
        model_config = SettingsConfigDict(
            env_exclude_inherit_parent=[
                "env_file_encoding",
                "env_exclude_inherit_parent",
            ],
        )

    class AppConfig(ArFiSettings):
        app: AppSettings
        model_config = SettingsConfigDict(
            env_prefix_as_source_mode_dir=True,
            env_nested_delimiter="---@---",
            env_file_encoding="cp1251",
            env_file=[".env", ".env.prod"],
        )

    config = AppConfig()
    assert config.app.city.address.settings_config.env_prefix == "app_city_address_"
    assert config.app.city.address.settings_config.env_file_encoding is None
    assert config.app.city.address.settings_config.env_nested_delimiter == ""
    assert config.app.city.address.settings_config.env_file == [".env", ".env.prod"]
    assert config.app.city.address.settings_config.env_exclude_inherit_parent == [
        "env_exclude_inherit_parent",
        "env_include_inherit_parent",
    ]
    assert config.app.city.address.settings_config.exclude_inherit_parent == [
        "env_exclude_inherit_parent",
        "env_include_inherit_parent",
    ]
    assert config.app.city.settings_config.env_prefix == "city_"
    assert config.app.city.settings_config.env_file_encoding is None
    assert config.app.city.settings_config.env_nested_delimiter == ""
    assert config.app.city.settings_config.env_file == [".env", ".env.prod"]
    assert config.app.city.settings_config.env_include_inherit_parent == []
    assert config.app.city.settings_config.env_exclude_inherit_parent == [
        "env_nested_delimiter",
        "env_exclude_inherit_parent",
        "env_prefix_as_mode_dir",
    ]
    assert config.app.city.settings_config.exclude_inherit_parent == [
        "env_nested_delimiter",
        "env_exclude_inherit_parent",
        "env_prefix_as_mode_dir",
    ]
    assert config.app.settings_config.env_prefix == "app_"
    assert config.app.settings_config.env_file_encoding is None
    assert config.app.settings_config.env_nested_delimiter == "---@---"
    assert config.app.settings_config.env_file == [".env", ".env.prod"]
    assert config.settings_config.env_prefix == ""
    assert config.settings_config.env_file_encoding == "cp1251"
    assert config.settings_config.env_nested_delimiter == "---@---"
    assert config.settings_config.env_file == [".env", ".env.prod"]


# @pytest.mark.current
@pytest.mark.env_config
def test_env_include_inherit_parent_exclude_self(cwd_to_tmp):
    class Proxy(ArFiSettings):
        file_config_inherit_parent = False
        model_config = SettingsConfigDict(
            env_prefix="proxy_test_",
            env_exclude_inherit_parent=[
                "env_file_encoding",
                "env_prefix",
            ],
        )

    class AppSettings(ArFiSettings):
        proxy: Proxy = Proxy()
        model_config = SettingsConfigDict(
            env_include_inherit_parent=[
                "env_file_encoding",
            ],
            env_exclude_inherit_parent=[
                "env_include_inherit_parent",
            ],
        )

    class AppConfig(ArFiSettings):
        app: AppSettings
        model_config = SettingsConfigDict(
            env_prefix="test_",
            env_file_encoding="cp1251",
        )

    config = AppConfig()
    assert config.settings_config.env_prefix == "test_"
    assert config.settings_config.env_file_encoding == "cp1251"
    assert config.app.settings_config.env_prefix == ""
    assert config.app.settings_config.env_file_encoding == "cp1251"
    assert config.app.proxy.settings_config.env_prefix == ""
    assert config.app.proxy.settings_config.env_file_encoding == "cp1251"
    assert config.app.proxy.settings_config.env_include_inherit_parent == []
    assert config.app.proxy.settings_config.env_exclude_inherit_parent == [
        "env_include_inherit_parent",
    ]
    assert config.app.proxy.inherited_params == [
        "env_exclude_inherit_parent",
        "env_file",
        "env_prefix",
        "env_prefix_as_mode_dir",
        "env_prefix_as_nested_mode_dir",
        "env_prefix_as_source_mode_dir",
        "env_file_encoding",
        "env_case_sensitive",
        "env_nested_delimiter",
        "env_ignore_missing",
    ]

    class Proxy(ArFiSettings):
        file_config_inherit_parent = False
        model_config = SettingsConfigDict(
            env_prefix="proxy_test_",
            env_exclude_inherit_parent=[
                "env_exclude_inherit_parent",
                "env_file_encoding",
                "env_prefix",
            ],
        )

    class AppSettings(ArFiSettings):
        proxy: Proxy = Proxy()
        model_config = SettingsConfigDict(
            env_include_inherit_parent=[
                "env_file_encoding",
            ],
            env_exclude_inherit_parent=[
                "env_include_inherit_parent",
            ],
        )

    class AppConfig(ArFiSettings):
        app: AppSettings
        model_config = SettingsConfigDict(
            env_prefix="test_",
            env_file_encoding="cp1251",
        )

    config = AppConfig()
    assert config.settings_config.env_prefix == "test_"
    assert config.settings_config.env_file_encoding == "cp1251"
    assert config.app.settings_config.env_prefix == ""
    assert config.app.settings_config.env_file_encoding == "cp1251"
    assert config.app.proxy.settings_config.env_prefix == "proxy_test_"
    assert config.app.proxy.settings_config.env_file_encoding is None
    assert config.app.proxy.settings_config.env_include_inherit_parent == []
    assert config.app.proxy.settings_config.env_exclude_inherit_parent == [
        "env_exclude_inherit_parent",
        "env_file_encoding",
        "env_prefix",
    ]
    assert config.app.proxy.inherited_params == [
        "env_include_inherit_parent",
        "env_file",
        "env_prefix_as_mode_dir",
        "env_prefix_as_nested_mode_dir",
        "env_prefix_as_source_mode_dir",
        "env_case_sensitive",
        "env_nested_delimiter",
        "env_ignore_missing",
    ]


# @pytest.mark.current
@pytest.mark.env_config
def test_env_prefix_as_mode_dir(cwd_to_tmp):
    class Nested(ArFiSettings):
        mode_dir = "nested"

    class Child(Nested):
        mode_dir = "child"

        env_config = EnvConfigDict(
            env_prefix_as_mode_dir=True,
            env_exclude_inherit_parent=[
                "env_prefix_as_mode_dir",
            ],
        )

    class AppSettings(ArFiSettings):
        child: Child

    class AppConfig(ArFiSettings):
        mode_dir = "dev"
        app: AppSettings

    config = AppConfig()
    assert config.app.mode_dir == "app"
    assert config.app.computed_mode_dir == "dev/app"
    assert config.app.child.mode_dir == "nested/child"
    assert config.app.child.computed_mode_dir == "dev/app/nested/child"
    assert config.app.child.settings_config.env_prefix == "dev_app_nested_child_"

    class Child(Nested):
        mode_dir = "child"

        env_config = EnvConfigDict(
            env_prefix_as_nested_mode_dir=True,
            env_exclude_inherit_parent=[
                "env_prefix_as_nested_mode_dir",
            ],
        )

    class AppSettings(ArFiSettings):
        child: Child
        env_config = EnvConfigDict(
            env_prefix_as_mode_dir=True,
        )

    class AppConfig(ArFiSettings):
        mode_dir = "dev"
        app: AppSettings

    config = AppConfig()
    assert config.app.child.settings_config.env_prefix == "nested_child_"

    class Child(Nested):
        mode_dir = "child"

        env_config = EnvConfigDict(
            env_prefix_as_source_mode_dir=True,
            env_exclude_inherit_parent=[
                "env_prefix_as_source_mode_dir",
            ],
        )

    class AppSettings(ArFiSettings):
        child: Child
        env_config = EnvConfigDict(
            env_prefix_as_mode_dir=True,
        )

    class AppConfig(ArFiSettings):
        mode_dir = "dev"
        app: AppSettings

    config = AppConfig()
    assert config.app.child.settings_config.env_prefix == "child_"

    class Child(Nested):
        mode_dir = "child"

        env_config = EnvConfigDict(
            env_prefix_as_mode_dir=True,
            env_exclude_inherit_parent=[
                "env_exclude_inherit_parent",
                "env_prefix_as_mode_dir",
                "env_prefix_as_nested_mode_dir",
                "env_prefix_as_source_mode_dir",
            ],
        )

    class AppSettings(ArFiSettings):
        child: Child
        env_config = EnvConfigDict(
            env_prefix_as_mode_dir=True,
            env_exclude_inherit_parent=[
                "env_prefix_as_mode_dir",
                "env_prefix_as_source_mode_dir",
            ],
        )

    class AppConfig(ArFiSettings):
        mode_dir = "dev"
        app: AppSettings
        env_config = EnvConfigDict(
            env_prefix_as_source_mode_dir=True,
        )

    config = AppConfig()
    assert config.settings_config.env_prefix == "dev_"
    assert config.app.settings_config.env_prefix == "dev_app_"
    assert config.app.child.settings_config.env_prefix == "dev_app_nested_child_"

    class Child(Nested):
        mode_dir = "child"

    class AppSettings(ArFiSettings):
        child: Child

    class AppConfig(ArFiSettings):
        mode_dir = "dev"
        app: AppSettings
        env_config = EnvConfigDict(
            env_prefix_as_mode_dir=True,
        )

    config = AppConfig()
    assert config.settings_config.env_prefix == "dev_"
    assert config.app.settings_config.env_prefix == "dev_app_"
    assert config.app.child.settings_config.env_prefix == "dev_app_nested_child_"


# @pytest.mark.current
@pytest.mark.env_config
@pytest.mark.parametrize("file_env", ["", None])
def test_env_config_disable_env_file(file_env):
    class AppConfig(ArFiSettings):
        env_config = EnvConfigDict(
            env_file=file_env,
        )

    config = AppConfig()
    assert config.computed_env_config["env_file"] == file_env
    assert config.env_path == []
    assert config.settings_config.env_file == file_env
    assert config.settings_config.env_path == []



import pytest

from arfi_settings.schemes import FileConfigSchema, SettingsParamsSchema, SettingsConfigSchema
from pathlib import PosixPath


# @pytest.mark.current
@pytest.mark.schemes
def test_file_config_schema():
    result_dict = FileConfigSchema().conf_dict
    assert result_dict
    assert isinstance(result_dict, dict)

    data = {"conf_ext": "toml, yaml, json,ini,conf"}
    result_dict = FileConfigSchema(**data).conf_dict
    assert result_dict["conf_ext"] == ["toml", "yaml", "json", "ini", "conf"]


# @pytest.mark.current
@pytest.mark.schemes
def test_settings_config_schema():
    result_dict = SettingsConfigSchema().conf_dict
    assert result_dict
    assert isinstance(result_dict, dict)

    data = {"conf_ext": "toml, yaml, json,ini,conf"}
    result_dict = SettingsConfigSchema(**data).conf_dict
    assert result_dict["conf_ext"] == ["toml", "yaml", "json", "ini", "conf"]

    settings = SettingsConfigSchema()
    assert settings.config_model_dict == {
        "case_sensitive": False,
        "ignore_missing": True,
        "encoding": None,
        "cli": False,
        "secrets_dir": None,
        "conf_file": "config",
        "conf_dir": "config",
        "conf_ext": ["toml", "yaml", "yml", "json"],
        "conf_file_encoding": None,
        "conf_case_sensitive": False,
        "conf_ignore_missing": True,
        "conf_custom_ext_handler": None,
        "env_file": ".env",
        "env_prefix": "",
        "env_prefix_as_mode_dir": False,
        "env_prefix_as_nested_mode_dir": False,
        "env_prefix_as_source_mode_dir": False,
        "env_file_encoding": None,
        "env_case_sensitive": False,
        "env_nested_delimiter": "",
        "env_ignore_missing": True,
        "conf_path": [],
        "env_path": [],
        "conf_include_inherit_parent": [],
        "conf_exclude_inherit_parent": [],
        "env_include_inherit_parent": [],
        "env_exclude_inherit_parent": [],
        "include_inherit_parent": [],
        "exclude_inherit_parent": [],
    }


# @pytest.mark.current
@pytest.mark.schemes
def test_settings_params_schema():
    result_dict = SettingsParamsSchema().default_param_dict
    assert isinstance(result_dict, dict)
    assert result_dict == {
        "read_config": True,
        "read_config_force": False,
        "mode_dir": PosixPath("."),
        "mode_dir_inherit_nested": True,
        "mode_dir_inherit_parent": True,
        "file_config_inherit_parent": True,
        "env_config_inherit_parent": True,
        "conf_file": "config",
        "conf_dir": "config",
        "conf_ext": ["toml", "yaml", "yml", "json"],
        "conf_file_encoding": None,
        "conf_case_sensitive": False,
        "conf_ignore_missing": True,
        "conf_custom_ext_handler": None,
        "conf_include_inherit_parent": [],
        "conf_exclude_inherit_parent": [],
        "env_file": ".env",
        "env_prefix": "",
        "env_prefix_as_mode_dir": False,
        "env_prefix_as_nested_mode_dir": False,
        "env_prefix_as_source_mode_dir": False,
        "env_file_encoding": None,
        "env_case_sensitive": False,
        "env_nested_delimiter": "",
        "env_ignore_missing": True,
        "env_include_inherit_parent": [],
        "env_exclude_inherit_parent": [],
        "case_sensitive": False,
        "ignore_missing": True,
        "encoding": None,
        "cli": False,
        "secrets_dir": None,
        "include_inherit_parent": [],
        "exclude_inherit_parent": [],
        "handler": "",
        "handler_inherit_parent": True,
        "ordered_settings": [
            "cli",
            "init_kwargs",
            "env",
            "env_file",
            "secrets",
            "conf_file",
        ],
        "ordered_settings_inherit_parent": True,
    }

    data = {"_conf_ext": "toml, yaml, json,ini,conf"}
    result_dict = SettingsParamsSchema(**data).default_param_dict
    assert result_dict["conf_ext"] == ["toml", "yaml", "json", "ini", "conf"]
    _kwargs = SettingsParamsSchema(**data).get_param_dict()
    assert _kwargs and isinstance(_kwargs, dict)
    assert _kwargs == {
        "_read_config": None,
        "_read_config_force": None,
        "_mode_dir": PosixPath("."),
        "_mode_dir_inherit_nested": None,
        "_mode_dir_inherit_parent": None,
        "_file_config_inherit_parent": None,
        "_env_config_inherit_parent": None,
        "_conf_file": PosixPath("."),
        "_conf_dir": PosixPath("."),
        "_conf_ext": ["toml", "yaml", "json", "ini", "conf"],
        "_conf_file_encoding": "",
        "_conf_case_sensitive": None,
        "_conf_ignore_missing": None,
        "_conf_custom_ext_handler": "",
        "_conf_include_inherit_parent": None,
        "_conf_exclude_inherit_parent": None,
        "_env_file": PosixPath("."),
        "_env_prefix": None,
        "_env_prefix_as_mode_dir": None,
        "_env_prefix_as_nested_mode_dir": None,
        "_env_prefix_as_source_mode_dir": None,
        "_env_file_encoding": "",
        "_env_case_sensitive": None,
        "_env_nested_delimiter": None,
        "_env_ignore_missing": None,
        "_env_include_inherit_parent": None,
        "_env_exclude_inherit_parent": None,
        "_case_sensitive": None,
        "_ignore_missing": None,
        "_encoding": "",
        "_cli": None,
        "_secrets_dir": PosixPath("."),
        "_include_inherit_parent": None,
        "_exclude_inherit_parent": None,
        "_handler": "",
        "_handler_inherit_parent": None,
        "_ordered_settings": [""],
        "_ordered_settings_inherit_parent": None,
    }

    params = SettingsParamsSchema(**data)
    assert params.default_param_dict["conf_ext"] == ["toml", "yaml", "json", "ini", "conf"]
    params.update_exclude_default(**{"_conf_ext": "toml,ini"})
    assert params.default_param_dict["conf_ext"] == ["toml", "ini"]



from pathlib import Path
from typing import Any, Literal

from pydantic import BaseModel, ConfigDict, Field, field_validator

from .constants import (
    CONFIG_EXT_DEFAULT,
    ORDERED_SETTINGS,
)
from .entity import (
    EnvConfigEntity as env,
    FileConfigEntity as conf,
)
from .types import (
    DEFAULT_PATH_SENTINEL,
    LIST_STR_SENTINEL,
    STR_SENTINEL,
    EnvConfigDict,
    FileConfigDict,
    MultiPathType,
    PathType,
    SettingsConfigDict,
    SettingsParamsDict,
)

__all__ = [
    "EnvConfigSchema",
    "FileConfigSchema",
    "SettingsConfigSchema",
    "SettingsParamsSchema",
    "PyProjectSchema",
    "CONF_INCLUDE_EXLUDE_PARAMS",
    "ENV_INCLUDE_EXLUDE_PARAMS",
    "INCLUDE_EXLUDE_PARAMS",
]

CONF_INCLUDE_EXLUDE_PARAMS = [
    "conf_file",
    "conf_dir",
    "conf_ext",
    "conf_file_encoding",
    "conf_case_sensitive",
    "conf_ignore_missing",
    "conf_custom_ext_handler",
    "conf_include_inherit_parent",
    "conf_exclude_inherit_parent",
]
ENV_INCLUDE_EXLUDE_PARAMS = [
    "env_file",
    "env_prefix",
    "env_prefix_as_mode_dir",
    "env_prefix_as_nested_mode_dir",
    "env_prefix_as_source_mode_dir",
    "env_file_encoding",
    "env_case_sensitive",
    "env_nested_delimiter",
    "env_ignore_missing",
    "env_include_inherit_parent",
    "env_exclude_inherit_parent",
]
INCLUDE_EXLUDE_PARAMS = []
INCLUDE_EXLUDE_PARAMS.extend(CONF_INCLUDE_EXLUDE_PARAMS)
INCLUDE_EXLUDE_PARAMS.extend(ENV_INCLUDE_EXLUDE_PARAMS)


class PyProjectSchema(BaseModel):
    """Default settings for each instance of the class ArFiSettings."""

    # arfi settings
    mode_dir_inherit_nested: bool | None = True
    mode_dir_inherit_parent: bool | None = True
    file_config_inherit_parent: bool | None = True
    env_config_inherit_parent: bool | None = True
    handler: str = "default_main_handler"
    handler_inherit_parent: bool = True
    ordered_settings: list[str] = ORDERED_SETTINGS
    ordered_settings_inherit_parent: bool = True

    # File settings
    conf_file: MultiPathType | None = conf.DEFAULT
    conf_dir: str | list[str] | None = conf.DEFAULT
    conf_ext: str | list[str] = CONFIG_EXT_DEFAULT
    conf_file_encoding: str | None = None
    conf_case_sensitive: bool | None = None
    conf_ignore_missing: bool | None = None
    conf_custom_ext_handler: str | dict[str, str] | None = None
    conf_include_inherit_parent: list[Literal[*CONF_INCLUDE_EXLUDE_PARAMS]] = []
    conf_exclude_inherit_parent: list[Literal[*CONF_INCLUDE_EXLUDE_PARAMS]] = []

    # Environment settings
    env_file: str | list[str] | None = ".env"
    env_prefix: str = ""
    env_prefix_as_mode_dir: bool = False
    env_prefix_as_nested_mode_dir: bool = False
    env_prefix_as_source_mode_dir: bool = False
    env_file_encoding: str | None = None
    env_case_sensitive: bool | None = None
    env_ignore_missing: bool | None = None
    env_nested_delimiter: str = ""
    env_include_inherit_parent: list[Literal[*ENV_INCLUDE_EXLUDE_PARAMS]] = []
    env_exclude_inherit_parent: list[Literal[*ENV_INCLUDE_EXLUDE_PARAMS]] = []

    # Global settings
    case_sensitive: bool = False
    ignore_missing: bool = True
    encoding: str | None = None
    cli: bool = False
    secrets_dir: str | None = None
    include_inherit_parent: list[Literal[*INCLUDE_EXLUDE_PARAMS]] = []
    exclude_inherit_parent: list[Literal[*INCLUDE_EXLUDE_PARAMS]] = []

    arfi_debug: bool = False

    model_config = ConfigDict(
        extra="forbid",
        validate_assignment=True,
        validate_default=False,
        validate_return=False,
    )


class UpdateMixin:
    def update(self, **new_data):
        for field, value in new_data.items():
            setattr(self, field, value)


class FileConfigSchema(BaseModel, UpdateMixin):
    """Validation schema for configuring file settings."""

    conf_file: MultiPathType | None = DEFAULT_PATH_SENTINEL
    conf_dir: MultiPathType | None = DEFAULT_PATH_SENTINEL
    conf_ext: str | list[str] = []
    conf_file_encoding: str | None = STR_SENTINEL
    conf_case_sensitive: bool | None = None
    conf_ignore_missing: bool | None = None
    conf_custom_ext_handler: str | dict[str, str] | None = STR_SENTINEL
    conf_include_inherit_parent: list[Literal[*CONF_INCLUDE_EXLUDE_PARAMS]] | None = None
    conf_exclude_inherit_parent: list[Literal[*CONF_INCLUDE_EXLUDE_PARAMS]] | None = None

    @field_validator("conf_ext")
    @classmethod
    def str_conf_ext(cls, v: str | list[str]) -> str | list[str]:
        """Ensure that the conf_ext is a string or list of strings."""
        if isinstance(v, str):
            v = [ext.strip() for ext in v.split(",")] or LIST_STR_SENTINEL
        return v

    model_config = ConfigDict(
        extra="forbid",
        validate_assignment=True,
        validate_default=False,
        validate_return=False,
    )

    @property
    def conf_dict(self) -> FileConfigDict:
        """Return a dictionary representation of the model."""
        return FileConfigDict(
            conf_file=self.conf_file if self.conf_file is not DEFAULT_PATH_SENTINEL else str(conf.DEFAULT),
            conf_dir=self.conf_dir if self.conf_dir is not DEFAULT_PATH_SENTINEL else str(conf.DEFAULT),
            conf_ext=self.conf_ext if self.conf_ext != [] else CONFIG_EXT_DEFAULT,
            conf_file_encoding=self.conf_file_encoding if self.conf_file_encoding is not STR_SENTINEL else None,
            conf_case_sensitive=self.conf_case_sensitive if self.conf_case_sensitive is not None else False,
            conf_ignore_missing=self.conf_ignore_missing if self.conf_ignore_missing is not None else True,
            conf_custom_ext_handler=(
                self.conf_custom_ext_handler if self.conf_custom_ext_handler is not STR_SENTINEL else None
            ),
            conf_include_inherit_parent=(
                self.conf_include_inherit_parent if self.conf_include_inherit_parent is not None else []
            ),
            conf_exclude_inherit_parent=(
                self.conf_exclude_inherit_parent if self.conf_exclude_inherit_parent is not None else []
            ),
        )


assert set(CONF_INCLUDE_EXLUDE_PARAMS) == set(FileConfigSchema.model_fields.keys())


class EnvConfigSchema(BaseModel, UpdateMixin):
    """Validation schema for configuring environment settings."""

    env_file: MultiPathType | None = DEFAULT_PATH_SENTINEL
    env_prefix: str | None = None
    env_prefix_as_mode_dir: bool | None = None
    env_prefix_as_nested_mode_dir: bool | None = None
    env_prefix_as_source_mode_dir: bool | None = None
    env_file_encoding: str | None = STR_SENTINEL
    env_case_sensitive: bool | None = None
    env_nested_delimiter: str | None = None
    env_ignore_missing: bool | None = None
    env_include_inherit_parent: list[Literal[*ENV_INCLUDE_EXLUDE_PARAMS]] | None = None
    env_exclude_inherit_parent: list[Literal[*ENV_INCLUDE_EXLUDE_PARAMS]] | None = None

    model_config = ConfigDict(
        extra="forbid",
        validate_assignment=True,
        validate_default=False,
        validate_return=False,
    )

    @property
    def env_dict(self) -> dict:
        """Return a dictionary representation of the model."""
        return EnvConfigDict(
            env_file=self.env_file if self.env_file is not DEFAULT_PATH_SENTINEL else env.DEFAULT_ENV_FILE,
            env_prefix=self.env_prefix if self.env_prefix is not None else "",
            env_prefix_as_mode_dir=self.env_prefix_as_mode_dir if self.env_prefix_as_mode_dir is not None else False,
            env_prefix_as_nested_mode_dir=(
                self.env_prefix_as_nested_mode_dir if self.env_prefix_as_nested_mode_dir is not None else False
            ),
            env_prefix_as_source_mode_dir=(
                self.env_prefix_as_source_mode_dir if self.env_prefix_as_source_mode_dir is not None else False
            ),
            env_file_encoding=self.env_file_encoding if self.env_file_encoding is not STR_SENTINEL else None,
            env_case_sensitive=self.env_case_sensitive if self.env_case_sensitive is not None else False,
            env_nested_delimiter=self.env_nested_delimiter if self.env_nested_delimiter is not None else "",
            env_ignore_missing=self.env_ignore_missing if self.env_ignore_missing is not None else True,
            env_include_inherit_parent=(
                self.env_include_inherit_parent if self.env_include_inherit_parent is not None else []
            ),
            env_exclude_inherit_parent=(
                self.env_exclude_inherit_parent if self.env_exclude_inherit_parent is not None else []
            ),
        )


assert set(ENV_INCLUDE_EXLUDE_PARAMS) == set(EnvConfigSchema.model_fields.keys())


class SettingsConfigSchema(FileConfigSchema, EnvConfigSchema):
    """Validation schema for configuring model settings."""

    case_sensitive: bool | None = None
    ignore_missing: bool | None = None
    encoding: str | None = STR_SENTINEL
    cli: bool | None = None
    secrets_dir: PathType | None = DEFAULT_PATH_SENTINEL
    conf_path: list[Path] = []
    env_path: list[Path] = []
    include_inherit_parent: list[Literal[*INCLUDE_EXLUDE_PARAMS]] | None = None
    exclude_inherit_parent: list[Literal[*INCLUDE_EXLUDE_PARAMS]] | None = None

    model_config = ConfigDict(
        extra="forbid",
        validate_assignment=True,
        validate_default=False,
        validate_return=False,
    )

    @property
    def config_model_dict(self) -> SettingsConfigDict:
        """Return a dictionary representation of the model."""
        settings_dict = SettingsConfigDict(
            case_sensitive=self.case_sensitive if self.case_sensitive is not None else False,
            ignore_missing=self.ignore_missing if self.ignore_missing is not None else True,
            encoding=self.encoding if self.encoding is not STR_SENTINEL else None,
            cli=self.cli if self.cli is not None else False,
            secrets_dir=self.secrets_dir if self.secrets_dir is not DEFAULT_PATH_SENTINEL else None,
            # secrets_dir=self.secrets_dir if self.secrets_dir != DEFAULT_PATH_SENTINEL else None,
            include_inherit_parent=self.include_inherit_parent if self.include_inherit_parent is not None else [],
            exclude_inherit_parent=self.exclude_inherit_parent if self.exclude_inherit_parent is not None else [],
        )
        settings_dict.update(self.conf_dict)
        settings_dict.update(self.env_dict)
        settings_dict.update(dict(conf_path=self.conf_path, env_path=self.env_path))
        return settings_dict


class SettingsParamsSchema(BaseModel):
    """Validation schema for init params."""

    # only params
    read_config: bool | None = Field(None, alias="_read_config")
    read_config_force: bool | None = Field(None, alias="_read_config_force")
    mode_dir: PathType | None = Field(DEFAULT_PATH_SENTINEL, alias="_mode_dir")
    mode_dir_inherit_nested: bool | None = Field(None, alias="_mode_dir_inherit_nested")
    mode_dir_inherit_parent: bool | None = Field(None, alias="_mode_dir_inherit_parent")
    file_config_inherit_parent: bool | None = Field(None, alias="_file_config_inherit_parent")
    env_config_inherit_parent: bool | None = Field(None, alias="_env_config_inherit_parent")

    # pyproject.toml params
    conf_file: MultiPathType | None = Field(DEFAULT_PATH_SENTINEL, alias="_conf_file")
    conf_dir: MultiPathType | None = Field(DEFAULT_PATH_SENTINEL, alias="_conf_dir")
    conf_ext: str | list[str] = Field([], alias="_conf_ext")
    conf_file_encoding: str | None = Field(STR_SENTINEL, alias="_conf_file_encoding")
    conf_case_sensitive: bool | None = Field(None, alias="_conf_case_sensitive")
    conf_ignore_missing: bool | None = Field(None, alias="_conf_ignore_missing")
    conf_custom_ext_handler: str | dict[str, str] | None = Field(STR_SENTINEL, alias="_conf_custom_ext_handler")
    conf_include_inherit_parent: list[Literal[*CONF_INCLUDE_EXLUDE_PARAMS]] | None = Field(
        None, alias="_conf_include_inherit_parent"
    )
    conf_exclude_inherit_parent: list[Literal[*CONF_INCLUDE_EXLUDE_PARAMS]] | None = Field(
        None, alias="_conf_exclude_inherit_parent"
    )
    env_file: MultiPathType | None = Field(DEFAULT_PATH_SENTINEL, alias="_env_file")
    env_prefix: str | None = Field(None, alias="_env_prefix")
    env_prefix_as_mode_dir: bool | None = Field(None, alias="_env_prefix_as_mode_dir")
    env_prefix_as_nested_mode_dir: bool | None = Field(None, alias="_env_prefix_as_nested_mode_dir")
    env_prefix_as_source_mode_dir: bool | None = Field(None, alias="_env_prefix_as_source_mode_dir")
    env_file_encoding: str | None = Field(STR_SENTINEL, alias="_env_file_encoding")
    env_case_sensitive: bool | None = Field(None, alias="_env_case_sensitive")
    env_nested_delimiter: str | None = Field(None, alias="_env_nested_delimiter")
    env_ignore_missing: bool | None = Field(None, alias="_env_ignore_missing")
    env_include_inherit_parent: list[Literal[*ENV_INCLUDE_EXLUDE_PARAMS]] | None = Field(
        None, alias="_env_include_inherit_parent"
    )
    env_exclude_inherit_parent: list[Literal[*ENV_INCLUDE_EXLUDE_PARAMS]] | None = Field(
        None, alias="_env_exclude_inherit_parent"
    )
    case_sensitive: bool | None = Field(None, alias="_case_sensitive")
    ignore_missing: bool | None = Field(None, alias="_ignore_missing")
    encoding: str | None = Field(STR_SENTINEL, alias="_encoding")
    cli: bool | None = Field(None, alias="_cli")
    secrets_dir: PathType | None = Field(DEFAULT_PATH_SENTINEL, alias="_secrets_dir")
    include_inherit_parent: list[Literal[*INCLUDE_EXLUDE_PARAMS]] | None = Field(None, alias="_include_inherit_parent")
    exclude_inherit_parent: list[Literal[*INCLUDE_EXLUDE_PARAMS]] | None = Field(None, alias="_exclude_inherit_parent")
    handler: str = Field(STR_SENTINEL, alias="_handler")
    handler_inherit_parent: bool | None = Field(None, alias="_handler_inherit_parent")
    ordered_settings: list[str] = Field(LIST_STR_SENTINEL, alias="_ordered_settings")
    ordered_settings_inherit_parent: bool | None = Field(None, alias="_ordered_settings_inherit_parent")

    model_config = ConfigDict(
        extra="ignore",
        validate_assignment=True,
        validate_default=False,
        validate_return=False,
    )

    @field_validator("conf_ext")
    @classmethod
    def str_conf_ext(cls, v: str | list[str] | None) -> str | list[str] | None:
        """Ensure that the conf_ext is a string or list of strings."""
        if isinstance(v, str):
            v = [ext.strip() for ext in v.split(",")] or LIST_STR_SENTINEL
        return v

    def get_param_dict(self, exclude_defaults=False) -> dict[str, Any]:
        params = self.model_dump(by_alias=True, exclude_defaults=exclude_defaults)
        return params

    @property
    def default_param_dict(self) -> SettingsParamsDict:
        return SettingsParamsDict(
            read_config=self.read_config if self.read_config is not None else True,
            read_config_force=self.read_config_force if self.read_config_force is not None else False,
            mode_dir=self.mode_dir,
            mode_dir_inherit_nested=self.mode_dir_inherit_nested if self.mode_dir_inherit_nested is not None else True,
            mode_dir_inherit_parent=self.mode_dir_inherit_parent if self.mode_dir_inherit_parent is not None else True,
            file_config_inherit_parent=(
                self.file_config_inherit_parent if self.file_config_inherit_parent is not None else True
            ),
            env_config_inherit_parent=(
                self.env_config_inherit_parent if self.env_config_inherit_parent is not None else True
            ),
            # file_config
            conf_file=self.conf_file if self.conf_file is not DEFAULT_PATH_SENTINEL else conf.DEFAULT,
            conf_dir=self.conf_dir if self.conf_dir is not DEFAULT_PATH_SENTINEL else conf.DEFAULT,
            conf_ext=self.conf_ext if self.conf_ext != [] else CONFIG_EXT_DEFAULT,
            conf_file_encoding=self.conf_file_encoding if self.conf_file_encoding is not STR_SENTINEL else None,
            conf_case_sensitive=self.conf_case_sensitive if self.conf_case_sensitive is not None else False,
            conf_ignore_missing=self.conf_ignore_missing if self.conf_ignore_missing is not None else True,
            conf_custom_ext_handler=(
                self.conf_custom_ext_handler if self.conf_custom_ext_handler is not STR_SENTINEL else None
            ),
            conf_include_inherit_parent=(
                self.conf_include_inherit_parent if self.conf_include_inherit_parent is not None else []
            ),
            conf_exclude_inherit_parent=(
                self.conf_exclude_inherit_parent if self.conf_exclude_inherit_parent is not None else []
            ),
            # env_config
            env_file=self.env_file if self.env_file is not DEFAULT_PATH_SENTINEL else env.DEFAULT_ENV_FILE,
            env_prefix=self.env_prefix if self.env_prefix is not None else "",
            env_prefix_as_mode_dir=self.env_prefix_as_mode_dir if self.env_prefix_as_mode_dir is not None else False,
            env_prefix_as_nested_mode_dir=(
                self.env_prefix_as_nested_mode_dir if self.env_prefix_as_nested_mode_dir is not None else False
            ),
            env_prefix_as_source_mode_dir=(
                self.env_prefix_as_source_mode_dir if self.env_prefix_as_source_mode_dir is not None else False
            ),
            env_file_encoding=self.env_file_encoding if self.env_file_encoding is not STR_SENTINEL else None,
            env_case_sensitive=self.env_case_sensitive if self.env_case_sensitive is not None else False,
            env_nested_delimiter=self.env_nested_delimiter if self.env_nested_delimiter is not None else "",
            env_ignore_missing=self.env_ignore_missing if self.env_ignore_missing is not None else True,
            env_include_inherit_parent=(
                self.env_include_inherit_parent if self.env_include_inherit_parent is not None else []
            ),
            env_exclude_inherit_parent=(
                self.env_exclude_inherit_parent if self.env_exclude_inherit_parent is not None else []
            ),
            # model_config
            case_sensitive=self.case_sensitive if self.case_sensitive is not None else False,
            ignore_missing=self.ignore_missing if self.ignore_missing is not None else True,
            encoding=self.encoding if self.encoding is not STR_SENTINEL else None,
            cli=self.cli if self.cli is not None else False,
            secrets_dir=self.secrets_dir if self.secrets_dir is not DEFAULT_PATH_SENTINEL else None,
            include_inherit_parent=self.include_inherit_parent if self.include_inherit_parent is not None else [],
            exclude_inherit_parent=self.exclude_inherit_parent if self.exclude_inherit_parent is not None else [],
            # handler settings
            handler=self.handler if self.handler is not STR_SENTINEL else "",
            handler_inherit_parent=self.handler_inherit_parent if self.handler_inherit_parent is not None else True,
            ordered_settings=self.ordered_settings if self.ordered_settings != LIST_STR_SENTINEL else ORDERED_SETTINGS,
            ordered_settings_inherit_parent=(
                self.ordered_settings_inherit_parent if self.ordered_settings_inherit_parent is not None else True
            ),
        )

    def update_exclude_default(self, **new_data):
        update_data = self.model_dump(by_alias=True)
        update_data.update(new_data)
        for key, value in self.model_validate(update_data).model_dump(exclude_defaults=True).items():
            setattr(self, key, value)
        return self



import sys

import pytest
from pytest_mock import MockFixture

from arfi_settings import ArFiBaseReader, ArFiReader, ArFiSettingsError

try:
    import yaml
except ImportError:
    yaml = None
try:
    import tomli
except ImportError:
    tomli = None


@pytest.fixture
def readers():
    from arfi_settings import readers

    readers.tomli = None
    readers.tomllib = None
    yield readers
    readers.tomli = None
    readers.tomllib = None


@pytest.fixture
def readers_tomli_is_not_none():
    from arfi_settings import readers

    readers.import_toml()
    yield readers
    readers.tomli = None
    readers.tomllib = None


@pytest.mark.readers
@pytest.mark.skipif(sys.version_info < (3, 11), reason="requires python3.11+")
def test_import_tomllib(readers):
    assert readers.tomli is None
    assert readers.tomllib is None
    readers.import_toml()
    assert readers.tomllib
    assert readers.tomli is None

    # readers.tomli = "tomli_test"
    # if readers.tomli is not None:
    #     assert readers.import_toml() is None
    #     assert readers.tomli


# @pytest.mark.skipif(sys.version_info >= (3, 11), reason="not need to use tomli")
@pytest.mark.readers
@pytest.mark.skipif(tomli is None, reason="tomli not installed")
def test_import_tomli_and_tomli_installed(mocker: MockFixture, readers):
    mock_sys = mocker.patch("arfi_settings.readers.sys")
    mock_sys.version_info = (3, 10)
    assert readers.tomli is None
    assert readers.tomllib is None
    readers.import_toml()
    assert readers.tomli
    assert readers.tomllib is None

    result = readers.import_toml()
    assert result is None


# @pytest.mark.skipif(sys.version_info >= (3, 11), reason="not need to use tomli")
@pytest.mark.readers
@pytest.mark.skipif(tomli, reason="tomli installed")
def test_import_tomli_and_tomli_not_installed(mocker: MockFixture, readers):
    mock_sys = mocker.patch("arfi_settings.readers.sys")
    mock_sys.version_info = (3, 10)
    assert readers.tomli is None
    assert readers.tomllib is None
    with pytest.raises(ArFiSettingsError) as excinfo:
        readers.import_toml()
    assert "tomli is not installed" in str(excinfo.value)

    readers.tomli = "tomli_test"
    result = readers.import_toml()
    assert result is None
    assert readers.tomli


@pytest.mark.readers
@pytest.mark.skipif(yaml is None, reason="PyYaml not installed")
def test_yaml_installed(mocker: MockFixture):
    from arfi_settings import readers

    # mocker.patch("arfi_settings.readers.yaml", return_value=None)
    # mocker.patch("readers.yaml", return_value=None)
    # assert readers.yaml is None
    readers.import_yaml()
    assert readers.yaml


@pytest.mark.readers
@pytest.mark.skipif(yaml, reason="PyYaml installed")
def test_yaml_not_installed(mocker: MockFixture):
    from arfi_settings import readers

    assert readers.yaml is None
    with pytest.raises(ArFiSettingsError) as excinfo:
        readers.import_yaml()
    assert "PyYaml not installed" in str(excinfo.value)


@pytest.mark.readers
def test_new_class_reader_without_method_read():
    with pytest.raises(TypeError):

        class DummyReader(ArFiBaseReader):
            def custom_reader(self):
                pass

        _ = DummyReader()


# @pytest.mark.current
@pytest.mark.readers
def test_new_class_reader():
    class DummyReader(ArFiReader):
        def read(self):
            pass

    _ = DummyReader()


@pytest.mark.readers
def test_read_without_params():
    with pytest.raises(ArFiSettingsError) as excinfo:
        reader = ArFiReader()
        reader.read()
    assert "No sources provided for the reading." in str(excinfo.value)


# @pytest.mark.current
@pytest.mark.readers
def test_read_toml(empty_config_config_toml):
    ArFiReader.BASE_DIR = None
    reader = ArFiReader(
        file_path="config/config.toml",
    )

    data = reader.read()
    assert data == {}


# @pytest.mark.current
@pytest.mark.readers
def test_read_data_toml(simple_data_config_config_toml):
    ArFiReader.BASE_DIR = None
    reader = ArFiReader(
        file_path="config/config.toml",
    )
    data = reader.read()
    assert data
    assert data == {"path_config_file": "config/config.toml"}


@pytest.mark.readers
def test_read_cnf_file_without_reader(simple_data_config_cnf):
    ArFiReader.BASE_DIR = None
    with pytest.raises(ArFiSettingsError) as excinfo:
        reader = ArFiReader(
            file_path="config.cnf",
        )
        reader.read()

    assert "`cnf_reader` is not defined" in str(excinfo.value)


# @pytest.mark.current
@pytest.mark.readers
def test_read_cnf_file_with_reader(simple_data_config_cnf):
    ArFiReader.BASE_DIR = None
    reader = ArFiReader(
        file_path="config.cnf",
        reader="toml",
    )
    data = reader.read()
    assert data
    assert data == {"path_config_file": "config.cnf"}


@pytest.mark.readers
def test_custom_reader(simple_data_config_cnf):
    ArFiReader.BASE_DIR = None

    class DummyReader(ArFiReader):
        def cnf_reader(self):
            return self.toml_reader()

    reader = DummyReader(
        file_path="config.cnf",
        reader="cnf",
    )
    data = reader.read()
    assert data
    assert data == {"path_config_file": "config.cnf"}


@pytest.mark.readers
def test_custom_reader_name_without_reader(simple_data_config_cnf):
    ArFiReader.BASE_DIR = None
    with pytest.raises(ArFiSettingsError) as excinfo:
        reader = ArFiReader(
            file_path="config.cnf",
            reader="custom",
        )
        reader.read()
    assert "`custom_reader` is not defined" in str(excinfo.value)


# @pytest.mark.current
@pytest.mark.readers
def test_ignore_missing(mocker: MockFixture):
    ArFiReader.BASE_DIR = None
    reader = ArFiReader(
        file_path="config.toml",
        ignore_missing=True,
    )
    data = reader.read()
    assert data == {}

    with pytest.raises(ArFiSettingsError) as excinfo:
        reader = ArFiReader(
            file_path="config.toml",
        )
        _ = reader.toml_reader()
    assert "File not found" in str(excinfo.value)

    mock_sys = mocker.patch("arfi_settings.readers.sys")
    mock_sys.version_info = (3, 10)
    reader = ArFiReader(
        file_path="config.toml",
        ignore_missing=True,
    )
    data = reader.toml_reader()
    assert data == {}

    with pytest.raises(ArFiSettingsError) as excinfo:
        reader = ArFiReader(
            file_path="config.toml",
        )
        _ = reader.toml_reader()
    assert "File not found" in str(excinfo.value)


# @pytest.mark.current
@pytest.mark.readers
def test_yaml_reader(simple_data_config_config_yaml):
    ArFiReader.BASE_DIR = None
    reader = ArFiReader(
        file_path="config/config.yaml",
    )
    data = reader.read()
    assert data
    assert data == {"path_config_file": "config/config.yaml"}


# @pytest.mark.current
@pytest.mark.readers
def test_non_exist_file_yaml_reader():
    with pytest.raises(ArFiSettingsError) as excinfo:
        reader = ArFiReader(
            file_path="config/config.yaml",
        )
        _ = reader.read()
    assert "File not found" in str(excinfo.value)

    reader = ArFiReader(
        file_path="config/config.yaml",
        ignore_missing=True,
    )
    data = reader.read()
    assert data == {}


@pytest.mark.readers
def test_yml_reader(simple_data_config_config_yml):
    ArFiReader.BASE_DIR = None
    reader = ArFiReader(
        file_path="config/config.yml",
    )
    data = reader.read()
    assert data
    assert data == {"path_config_file": "config/config.yml"}


@pytest.mark.readers
def test_non_exist_file_yml_reader():
    with pytest.raises(ArFiSettingsError) as excinfo:
        reader = ArFiReader(
            file_path="config/config.yml",
        )
        _ = reader.read()

    assert "File not found" in str(excinfo.value)

    reader = ArFiReader(
        file_path="config/config.yml",
        ignore_missing=True,
    )
    data = reader.read()
    assert data == {}


@pytest.mark.readers
def test_json_reader(simple_data_config_config_json):
    ArFiReader.BASE_DIR = None
    reader = ArFiReader(
        file_path="config/config.json",
    )
    data = reader.read()
    assert data
    assert data == {"path_config_file": "config/config.json"}


@pytest.mark.readers
def test_non_exist_file_json_reader():
    with pytest.raises(ArFiSettingsError) as excinfo:
        reader = ArFiReader(
            file_path="config/config.json",
        )
        _ = reader.read()

    assert "File not found" in str(excinfo.value)

    reader = ArFiReader(
        file_path="config/config.json",
        ignore_missing=True,
    )
    data = reader.read()
    assert data == {}


# @pytest.mark.current
@pytest.mark.readers
def test_read_env_file_non_exist(cwd_to_tmp, path_base_dir):
    with pytest.raises(ArFiSettingsError) as excinfo:
        reader = ArFiReader(
            file_path=".env",
            is_env_file=True,
        )
        _ = reader.read()
    assert "File not found" in str(excinfo.value)

    reader = ArFiReader(
        file_path=".env",
        is_env_file=True,
        ignore_missing=True,
    )
    data = reader.read()
    assert data == {}


@pytest.mark.readers
def test_read_env_file(simple_data_env_file):
    ArFiReader.BASE_DIR = None
    reader = ArFiReader(
        file_path=".env",
        is_env_file=True,
    )
    data = reader.read()
    assert data
    assert data == {"PATH_CONFIG_FILE": ".env"}


@pytest.mark.readers
def test_read_env(monkeypatch):
    monkeypatch.setenv("PATH_CONFIG_FILE", ".env.prod")
    reader = ArFiReader(
        is_env=True,
    )
    data = reader.read()
    assert data
    assert data["PATH_CONFIG_FILE"] == ".env.prod"


@pytest.mark.readers
@pytest.mark.parametrize(
    "reader_name, error_message",
    [
        (12, "Reader `name` must be not empty string"),
        ("_test_reader", "Reader `name` must not start with `_`"),
    ],
)
def test_normolize_reader_name(reader_name, error_message):
    with pytest.raises(ArFiSettingsError) as excinfo:
        reader = ArFiReader(
            file_path=".env",
            is_env_file=True,
            reader=reader_name,
        )
        _ = reader.read()
    assert error_message in str(excinfo.value)


# @pytest.mark.current
@pytest.mark.readers
@pytest.mark.parametrize(
    "bad_file_path",
    ["", 12, ".", "./", "/", "\\\\", {"path"}],
)
def test_validate_file_path(bad_file_path):
    with pytest.raises(ArFiSettingsError) as excinfo:
        reader = ArFiReader(
            file_path=bad_file_path,
        )
        _ = reader.read()
    assert "File path is not set or invalid" in str(excinfo.value)


# @pytest.mark.current
@pytest.mark.readers
@pytest.mark.parametrize(
    "file_path, resp",
    [
        (123, False),
        ("config/config.toml", True),
        ("config", False),
        (None, False),
        ("config/config.yaml", False),
        ({"asd", 123}, False),
    ],
)
def test_is_exist_file(empty_config_config_toml, file_path, resp):
    ArFiReader.BASE_DIR = None
    result = ArFiReader.is_exist_file(file_path)
    assert result == resp


# @pytest.mark.current
@pytest.mark.readers
def test_empty_param():
    with pytest.raises(ArFiSettingsError) as excinfo:
        reader = ArFiReader(ignore_missing=False)
        _ = reader.read()
    assert "No sources provided for the reading" in str(excinfo.value)


# @pytest.mark.current
@pytest.mark.readers
def test_file_without_ext_and_without_reader():
    with pytest.raises(ArFiSettingsError) as excinfo:
        reader = ArFiReader(file_path="config", ignore_missing=True)
        _ = reader.read()
    assert "No reader provided for file" in str(excinfo.value)


# @pytest.mark.current
@pytest.mark.secret
@pytest.mark.readers
def test_non_exist_secret_file(secrets_dir):
    ArFiReader.BASE_DIR = None
    reader = ArFiReader(
        file_path="secrets/path_config_file",
        ignore_missing=True,
        is_secret_file=True,
    )
    data = reader.read()
    assert data == {}

    with pytest.raises(ArFiSettingsError) as excinfo:
        reader = ArFiReader(
            file_path="secrets/path_config_file",
            is_secret_file=True,
        )
        _ = reader.read()
    assert "File not found" in str(excinfo.value)


# @pytest.mark.current
@pytest.mark.secret
@pytest.mark.readers
def test_empty_secret_file(empty_secret_path_config_file):
    ArFiReader.BASE_DIR = None
    reader = ArFiReader(
        file_path="secrets/path_config_file",
        is_secret_file=True,
    )
    data = reader.read()
    assert data == {"path_config_file": ""}


# @pytest.mark.current
@pytest.mark.secret
@pytest.mark.readers
def test_secret_file_with_data(simple_data_secret_path_config_file):
    ArFiReader.BASE_DIR = None
    reader = ArFiReader(
        file_path="secrets/path_config_file",
        is_secret_file=True,
    )
    data = reader.read()
    assert data == {"path_config_file": "secrets/path_config_file"}


# @pytest.mark.current
@pytest.mark.cli
@pytest.mark.readers
def test_setup_cli_reader():
    def parse_args():
        return ["test_key", "test_value"]

    with pytest.raises(ArFiSettingsError) as excinfo:
        ArFiReader.setup_cli_reader(parse_args)
        reader = ArFiReader(is_cli=True)
        _ = reader.read()
    assert "CLI reader must return a dict" in str(excinfo.value)

    def parse_args():
        return {"test_key": "test_value"}

    with pytest.raises(ArFiSettingsError) as excinfo:
        cli_reader = parse_args()
        ArFiReader.setup_cli_reader(cli_reader)
    assert "`cli_reader` must be callable" in str(excinfo.value)

    ArFiReader.setup_cli_reader(parse_args)
    reader = ArFiReader(is_cli=True)
    data = reader.read()
    assert data == {"test_key": "test_value"}

    def parse_args(*args):
        return {"test_key": "test_value"}

    ArFiReader.setup_cli_reader(parse_args)
    reader = ArFiReader(is_cli=True)
    data = reader.read()
    assert data == {"test_key": "test_value"}

    def parse_args(*args, **kwargs):
        return {"test_key": "test_value"}

    ArFiReader.setup_cli_reader(parse_args)
    reader = ArFiReader(is_cli=True)
    data = reader.read()
    assert data == {"test_key": "test_value"}

    def parse_args(test_value=25, asd=38):
        assert test_value == 25
        assert asd == 38
        return {"test_key": "test_value"}

    ArFiReader.setup_cli_reader(parse_args)
    reader = ArFiReader(is_cli=True)
    data = reader.read()
    assert data == {"test_key": "test_value"}

    class ParseArgs:
        def __call__(self):
            return {"test_key": "test_value"}

    cli_reader = ParseArgs()
    ArFiReader.setup_cli_reader(cli_reader)
    assert isinstance(ArFiReader.default_cli_reader, ParseArgs)
    reader = ArFiReader(is_cli=True)
    data = reader.read()
    assert data == {"test_key": "test_value"}


# @pytest.mark.current
@pytest.mark.cli
@pytest.mark.readers
def test_reader_whithout_default(monkeypatch):
    monkeypatch.setattr(ArFiReader, "default_cli_reader", None)
    reader = ArFiReader(is_cli=True)
    data = reader.read()
    assert data == {}



from pathlib import Path

import pytest

import arfi_settings
from arfi_settings import ArFiSettings
from arfi_settings.init_config import InitSettings


# @pytest.mark.current
@pytest.mark.settings
def test_base_dir(cwd_to_tmp, monkeypatch):
    class PatchedInitSettings(InitSettings):
        @staticmethod
        def _search_called_file():
            return cwd_to_tmp.as_posix(), 0

    monkeypatch.setattr(arfi_settings.init_config, "InitSettings", PatchedInitSettings)
    init_settings = PatchedInitSettings()
    monkeypatch.setattr(arfi_settings.init_config, "init_settings", init_settings)
    monkeypatch.setattr(arfi_settings.main, "init_settings", init_settings)

    class AppConfig(ArFiSettings):
        pass

    config = AppConfig()
    assert config.BASE_DIR == cwd_to_tmp.parent

    monkeypatch.setattr(arfi_settings.init_config, "InitSettings", PatchedInitSettings)
    init_settings = PatchedInitSettings()
    monkeypatch.setattr(arfi_settings.init_config, "init_settings", init_settings)
    monkeypatch.setattr(arfi_settings.main, "init_settings", init_settings)

    class AppConfig(ArFiSettings):
        BASE_DIR = "settings/"

    config = AppConfig()
    assert config.BASE_DIR == cwd_to_tmp / "settings"

    monkeypatch.setattr(arfi_settings.init_config, "InitSettings", PatchedInitSettings)
    init_settings = PatchedInitSettings()
    monkeypatch.setattr(arfi_settings.init_config, "init_settings", init_settings)
    monkeypatch.setattr(arfi_settings.main, "init_settings", init_settings)

    class AppConfig(ArFiSettings):
        BASE_DIR = "/settings"

    config = AppConfig()
    assert config.BASE_DIR == Path("/settings")

    monkeypatch.setattr(arfi_settings.init_config, "InitSettings", PatchedInitSettings)
    init_settings = PatchedInitSettings()
    monkeypatch.setattr(arfi_settings.init_config, "init_settings", init_settings)
    monkeypatch.setattr(arfi_settings.main, "init_settings", init_settings)
    monkeypatch.setenv("HOME", "/")

    class AppConfig(ArFiSettings):
        BASE_DIR = "~/settings"

    config = AppConfig()
    assert config.BASE_DIR == Path("/settings")


# @pytest.mark.settings
# def test_simple_config_file_by_default(simple_data_config_config_toml):
#     class AppConfig(ArFiSettings):
#         path_config_file: str
#
#     config = AppConfig()
#     assert config.MODE is None
#     assert config.path_config_file == "config/config.toml"
#
#
# # @pytest.mark.skip
# @pytest.mark.settings
# class TestMainConfig:
#     # @pytest.mark.current
#     def test_read_config_default_detection(self):
#         class AppConfig(ArFiSettings):
#             pass
#
#         assert AppConfig.read_config is False
#         config = AppConfig()
#         assert config.read_config is True
#
#     # @pytest.mark.current
#     def test_read_config_handler_setup(self):
#         class ProdAppSettings(ArFiSettings):
#             pass
#
#         class AppSettings(ArFiSettings):
#             asd: str = "test"
#             pass
#
#         class AppConfig(ArFiSettings):
#             app: AppSettings
#             # app: AppSettings = {}
#             # app: AppSettings = AppSettings()
#
#         assert AppSettings.read_config is False
#         assert ProdAppSettings.read_config is False
#         assert AppConfig.read_config is False
#
#         data = {
#             "app": {
#                 "asd": "asd",
#                 # "_read_config": True,
#             }
#         }
#         config = AppConfig(**data)
#         assert config.read_config is True
#         assert config.app.read_config is True
#
#         print("=" * 100)
#
#         class AppConfig(ArFiSettings):
#             app: AppSettings | ProdAppSettings = AppSettings()
#
#         config = AppConfig()
#         assert config.app.read_config is True
#
#         print("=" * 100)
#
#         class AppConfig(ArFiSettings):
#             app: AppSettings | None = AppSettings()
#
#         config = AppConfig()
#         assert config.app.read_config is False
#
#         print("=" * 100)
#
#         class AppConfig(ArFiSettings):
#             # app: ArFiSettings = None
#             app: AppSettings = None
#
#         config = AppConfig()
#         assert config.app.read_config is True
#
#
# # @pytest.mark.settings
# # class TestAliasesSettings:
# #     """Test aliases in settings."""
# #
# #     # @pytest.mark.current
# #     def test_without_alias(self, empty_config_config_toml):
# #         conf_file = empty_config_config_toml
# #         conf_file.write_text(
# #             """
# #             path_config_file = 'config/config.toml'
# #             [app]
# #             path_config_file = 'config/config.toml'
# #             """
# #         )
# #
# #         class AppSettings(ArFiSettings):
# #             path_config_file: str = "tests/"
# #
# #         class AppConfig(ArFiSettings):
# #             path_config_file: str
# #             app: AppSettings
# #
# #         config = AppConfig()
# #         assert config.MODE is None
# #         assert config.path_config_file == "config/config.toml"
# #         assert config.app.path_config_file == "config/config.toml"
# #
# #     # @pytest.mark.current
# #     def test_simple_alias(self, empty_config_config_toml):
# #         conf_file = empty_config_config_toml
# #         conf_file.write_text(
# #             """
# #             app_config_file = 'config/config.toml'
# #             [app]
# #             app_settings_file = 'config/config.toml'
# #             """
# #         )
# #
# #         class AppSettings(ArFiSettings):
# #             path_config_file: str = Field(alias="app_settings_file")
# #
# #         class AppConfig(ArFiSettings):
# #             path_config_file: str = Field(alias="app_config_file")
# #             app: AppSettings
# #
# #         config = AppConfig()
# #         assert config.MODE is None
# #         assert config.path_config_file == "config/config.toml"
# #         assert config.app.path_config_file == "config/config.toml"
# #
# #     # # @pytest.mark.current
# #     # def test_alias_case_sensitive_by_default(self, empty_config_config_toml):
# #     #     conf_file = empty_config_config_toml
# #     #     conf_file.write_text(
# #     #         """
# #     #         APP_CONFIG_FILE = 'config/config.toml'
# #     #         [app]
# #     #         APP_SETTINGS_FILE = 'config/config.toml'
# #     #         """
# #     #     )
# #     #
# #     #     class AppSettings(ArFiSettings):
# #     #         path_config_file: str = Field(alias="app_settings_file")
# #     #
# #     #     class AppConfig(ArFiSettings):
# #     #         path_config_file: str = Field(alias="app_config_file")
# #     #         app: AppSettings
# #     #
# #     #     config = AppConfig()
# #     #     assert config.MODE is None
# #     #     assert config.path_config_file == "config/config.toml"
# #     #     assert config.app.path_config_file == "config/config.toml"
#
#
# # @pytest.mark.current
# # @pytest.mark.settings
# # def test_explicit_config_file_by_default(simple_data_config_config_toml):
# #     from typing import Optional, Literal
# #     from pydantic import BaseModel
# #
# #     class ProdAppSettings(ArFiSettings):
# #         path_config_file: str = "prod/"
# #
# #     class AppSettings(ArFiSettings):
# #         path_config_file: str = "tests/"
# #
# #     class Simple(BaseModel):
# #         path_config_file: str = "base/"
# #
# #     class AppConfig(ArFiSettings):
# #         path_config_file: str
# #         # app: AppSettings
# #         app: AppSettings | ProdAppSettings
# #         # app: dict | AppSettings
# #         # app: AppSettings | dict
# #         # app: AppSettings | str | None
# #         # app: Optional[AppSettings]
# #         # app: Optional[AppSettings] | str | dict
# #         # app: AppSettings = AppSettings()
# #         # app: AppSettings = None
# #         # app: AppSettings | str | None | list | dict | Simple | ProdAppSettings = None
# #         # app: AppSettings | list[AppSettings] | str | None
# #         # app: list[AppSettings] | AppSettings | str | None
# #         # app: list[AppSettings] | AppSettings | str | None | Simple
# #
# #         # app: Literal["ArFiSettings", "ProdAppSettings", ArFiSettings()] | str
# #
# #     data = {
# #         "app": {"path_config_file": "config/config.toml"},
# #     }
# #     # data = {"app": AppSettings}
# #     config = AppConfig(**data)
# #     assert config.MODE is None
# #     assert config.path_config_file == "config/config.toml"
# #     assert config.app.path_config_file == "config/config.toml"
#
#
# # @pytest.mark.current
# # @pytest.mark.settings
# # def test_explicit_config_file_by_default2(tests_dir):
# #     print()
# #     import os
# #
# #     print(os.getcwd())
# #
# #     class Proxy(ArFiSettings):
# #         path_config_file: str
# #
# #     class AppSettings(ArFiSettings):
# #         path_config_file: str
# #         proxy: Proxy
# #
# #     class AppConfig(ArFiSettings):
# #         path_config_file: str
# #         app: AppSettings
# #
# #     config = AppConfig()
# #     assert config.MODE is None
# #     assert config.path_config_file == "config/config.toml"
# #     print()
# #     print(config.model_dump_json(indent=4))
#
#
# # # @pytest.mark.current
# # @pytest.mark.settings
# # def test_explicit_config_file_by_default_init(tests_dir):
# #     print()
# #     import os
# #
# #     print(os.getcwd())
# #
# #     class Proxy(ArFiSettings):
# #         path_config_file: str
# #
# #     class AppSettings(ArFiSettings):
# #         path_config_file: str
# #         proxy: Proxy
# #
# #     class AppConfig(ArFiSettings):
# #         path_config_file: str
# #         app: AppSettings = AppSettings()
# #
# #     config = AppConfig()
# #     assert config.MODE is None
# #     assert config.path_config_file == "config/config.toml"



import json
import os
import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable

from dotenv import dotenv_values

from .errors import ArFiSettingsError
from .types import PathType
from .utils import validate_cli_reader

if TYPE_CHECKING:
    if sys.version_info >= (3, 11):
        import tomllib
    else:
        tomllib = None
    import tomli
    import yaml
else:
    yaml = None
    tomllib = None
    tomli = None

__all__ = (
    "ArFiBaseReader",
    "ArFiReader",
)


def import_yaml() -> None:
    global yaml
    if yaml is not None:
        return
    try:
        import yaml
    except ImportError as e:
        raise ArFiSettingsError("PyYaml is not installed, run `pip install arfi-settings[yaml]`") from e


def import_toml() -> None:
    global tomli
    global tomllib
    if sys.version_info < (3, 11):
        if tomli is not None:
            return
        try:
            import tomli
        except ImportError as e:
            raise ArFiSettingsError("tomli is not installed, run `pip install arfi-settings[toml]`") from e
    else:
        if tomllib is not None:
            return
        import tomllib


class ArFiBaseReader(ABC):
    """Readers of the source settings."""

    default_cli_reader: Callable | None = None
    ROOT_DIR: PathType | None = None
    BASE_DIR: PathType | None = None

    def __init__(
        self,
        file_path: PathType | None = None,
        file_encoding: str | None = None,
        is_env_file: bool = False,
        is_env: bool = False,
        is_cli: bool = False,
        is_secret_file: bool = False,
        reader: str = "",
        ignore_missing: bool = False,
        **options,
    ) -> None:
        self.file_path = self._validate_file_path(file_path)
        self.file_encoding = file_encoding
        self.is_env_file = is_env_file
        self.is_env = is_env
        self.is_cli = is_cli
        self.is_secret_file = is_secret_file
        self.ignore_missing = ignore_missing
        self.reader = reader
        if self.reader:
            self.reader = self._normalize_reader_name(self.reader)
        self.options = options
        self._resolve_file_path()

    def _resolve_file_path(self):
        """Resolves file path."""

        if self.file_path is None:
            return
        if self.BASE_DIR is not None:
            self.BASE_DIR = Path(self.BASE_DIR).resolve()
        if self.ROOT_DIR is not None:
            self.ROOT_DIR = Path(self.ROOT_DIR).resolve()
        if not Path(self.file_path).is_absolute():
            if self.is_env_file and self.ROOT_DIR is not None:
                env_file_resolve = Path(self.ROOT_DIR) / self.file_path
                if env_file_resolve.is_file():
                    self.file_path = env_file_resolve
                else:
                    if self.BASE_DIR is not None:
                        self.file_path = Path(self.BASE_DIR) / self.file_path
            else:
                if self.BASE_DIR is not None:
                    self.file_path = Path(self.BASE_DIR) / self.file_path

    @staticmethod
    def _normalize_reader_name(name: str) -> str:
        """Normalizesza reader name."""

        try:
            assert name and isinstance(name, str), "Reader `name` must be not empty string"
            assert not name.startswith("_"), "Reader `name` must not start with `_`"
        except AssertionError as e:
            raise ArFiSettingsError(e) from e
        suffix = "reader"
        if not name.endswith(suffix):
            name = f"{name}_{suffix}"
        return name

    @classmethod
    def _is_exist_reader(cls, name: str) -> bool:
        """Check is exist reader."""

        name = cls._normalize_reader_name(name)
        return hasattr(cls, name)

    def _get_reader(self, name: str) -> Callable:
        """Get reader."""

        name = self._normalize_reader_name(name)
        if not self._is_exist_reader(name):
            raise ArFiSettingsError(f"{self.__class__.__name__}: `{name}` is not defined")
        return self.__getattribute__(name)

    @staticmethod
    def _validate_file_path(file_path: PathType | None) -> Path | None:
        """Validate and return file path."""

        if file_path is not None:
            try:
                if isinstance(file_path, str):
                    assert file_path
                if not isinstance(file_path, Path):
                    file_path = Path(file_path).expanduser()
                assert Path(file_path).as_posix().strip(".").strip("\\").strip("/")
            except (AssertionError, TypeError) as e:
                raise ArFiSettingsError(f"File path is not set or invalid: `{file_path}`") from e
            else:
                return file_path.expanduser()
        return file_path

    @classmethod
    def is_exist_file(cls, file_path: PathType | None) -> bool:
        """Checks if the file path is specified and if the file exists."""

        try:
            file_path: Path | None = cls._validate_file_path(file_path)
            if file_path is None:
                return False
        except ArFiSettingsError:
            return False
        if not file_path.exists():
            return False
        if not file_path.is_file():
            return False
        return True

    def toml_reader(self) -> dict[str, Any]:
        """Reads settings from TOML file."""

        import_toml()
        try:
            with open(self.file_path, mode="rb") as toml_file:
                if sys.version_info < (3, 11):
                    return tomli.load(toml_file)
                return tomllib.load(toml_file)
        except FileNotFoundError as e:
            if self.ignore_missing:
                return {}
            raise ArFiSettingsError(f"File not found: `{self.file_path.resolve().as_posix()}`") from e

    def yaml_reader(self) -> dict[str, Any]:
        """Reads settings from YAML file."""

        import_yaml()
        try:
            with open(self.file_path, encoding=self.file_encoding) as yaml_file:
                return yaml.safe_load(yaml_file)
        except FileNotFoundError as e:
            if self.ignore_missing:
                return {}
            raise ArFiSettingsError(f"File not found: `{self.file_path.resolve().as_posix()}`") from e

    def yml_reader(self) -> dict[str, Any]:
        """Reads settings from YAML file."""

        return self.yaml_reader()

    def json_reader(self) -> dict[str, Any]:
        """Reads settings from JSON file."""

        try:
            with open(self.file_path, encoding=self.file_encoding) as json_file:
                return json.load(json_file)
        except FileNotFoundError as e:
            if self.ignore_missing:
                return {}
            raise ArFiSettingsError(f"File not found: `{self.file_path.resolve().as_posix()}`") from e

    def env_file_reader(self) -> dict[str, str | None]:
        """Reads settings from .env file."""

        if not self.ignore_missing:
            if not self.is_exist_file(self.file_path):
                raise ArFiSettingsError(f"File not found: `{self.file_path.resolve().as_posix()}`")
        data: dict[str, str | None] = dotenv_values(self.file_path, encoding=self.file_encoding)
        return dict(data)

    def env_reader(self) -> dict[str, str | None]:
        """Reads settings from environment."""
        return dict(os.environ)

    def secret_file_reader(self) -> dict[str, str]:
        """Reads settings from secrets directory."""

        data = {}
        if not self.is_exist_file(self.file_path):
            if not self.ignore_missing:
                raise ArFiSettingsError(f"File not found: `{self.file_path.resolve().as_posix()}`")
            return data

        key_name = self.file_path.stem
        data[key_name] = self.file_path.read_text(encoding=self.file_encoding).strip()
        return data

    def cli_reader(self) -> dict[str, Any]:
        """Reads settings from CLI."""

        if self.default_cli_reader is not None:
            data = self.default_cli_reader()
            try:
                assert isinstance(data, dict)
            except AssertionError as e:
                raise ArFiSettingsError(f"CLI reader must return a dict. Got: {type(data)}") from e
            return data
        return {}

    @classmethod
    def setup_cli_reader(cls, cli_reader: Callable) -> None:
        """Setup default CLI reader.

        cli_reader: Callable Class or Function that returns a dict[srt, Any]
        """
        cls.default_cli_reader = validate_cli_reader(cli_reader)

    @abstractmethod
    def read(self) -> dict[str, Any]:
        """Read and return settings."""


class ArFiReader(ArFiBaseReader):
    """Reads settings from file."""

    def read(self) -> dict[str, Any]:
        if self.reader:
            reader = self._get_reader(self.reader)
            return reader()
        elif self.is_cli:
            return self.cli_reader()
        elif self.is_env:
            return self.env_reader()
        elif self.is_secret_file:
            return self.secret_file_reader()
        elif self.file_path is not None:
            if self.is_env_file:
                return self.env_file_reader()
            reader_name = self.file_path.suffix.strip(".")
            if not reader_name and not self.reader:
                raise ArFiSettingsError(f"No reader provided for file `{self.file_path.resolve().as_posix()}`.")
            reader_name = self._normalize_reader_name(reader_name or self.reader)
            if not self._is_exist_reader(reader_name):
                raise ArFiSettingsError(
                    f"{self.__class__.__name__}: reader `{reader_name}` is not defined "
                    f"for file `{self.file_path.resolve().as_posix()}`."
                )
            reader = self._get_reader(reader_name)
            return reader()
        else:
            raise ArFiSettingsError("No sources provided for the reading.")



from typing import Literal

import pytest
from pydantic import AliasChoices, AliasPath, Field

from arfi_settings import ArFiSettings, EnvConfigDict, FileConfigDict, SettingsConfigDict


# @pytest.mark.current
@pytest.mark.case_sensitive
def test_case_sensitive(alias_data_config_config_toml, path_base_dir):
    class AppConfig(ArFiSettings):
        pass

    config = AppConfig()
    assert config.settings_config.case_sensitive is False
    assert config.settings_config.conf_case_sensitive is False
    assert config.settings_config.env_case_sensitive is False

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(case_sensitive=True)

    config = AppConfig()
    assert config.settings_config.case_sensitive is True
    assert config.settings_config.conf_case_sensitive is True
    assert config.settings_config.env_case_sensitive is True

    class AppConfig(ArFiSettings):
        model_config = SettingsConfigDict(
            case_sensitive=True,
            conf_case_sensitive=False,
            env_case_sensitive=False,
        )

    config = AppConfig()
    assert config.settings_config.case_sensitive is True
    assert config.settings_config.conf_case_sensitive is False
    assert config.settings_config.env_case_sensitive is False

    class AppConfig(ArFiSettings):
        file_config = FileConfigDict(conf_case_sensitive=False)
        env_config = EnvConfigDict(env_case_sensitive=False)
        model_config = SettingsConfigDict(
            case_sensitive=True,
            conf_case_sensitive=True,
            env_case_sensitive=True,
        )

    config = AppConfig()
    assert config.settings_config.case_sensitive is True
    assert config.settings_config.conf_case_sensitive is False
    assert config.settings_config.env_case_sensitive is False

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(
            validation_alias=AliasChoices(
                AliasPath("ALIAS_PATH", 0),
                AliasPath("OTHER_ALIAS_PATH", 1),
            ),
        )
        asd: str = Field(validation_alias=AliasChoices("ASD_TEST", "CONFIG_DATA"))

    config = AppConfig()
    assert config.alias_config_data == "alias_path_1"
    assert config.asd == "asd_test"

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(
            validation_alias=AliasChoices(
                AliasPath("ALIAS_PATH", 0),
                AliasPath("OTHER_ALIAS_PATH", 1),
            ),
        )
        asd: str = Field(validation_alias=AliasChoices("ASD_TEST", "CONFIG_DATA"))
        model_config = SettingsConfigDict(case_sensitive=True)

    config = AppConfig()
    assert config.alias_config_data == "other_alias_path_2"
    assert config.asd == "CONFIG_DATA"

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(
            validation_alias=AliasChoices(
                AliasPath("other_complex_alias_path", "ALIAS_PATH", 1),
            ),
        )

    config = AppConfig()
    assert config.alias_config_data == "OTHER_ALIAS_PATH_2"

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(
            validation_alias=AliasChoices(
                AliasPath("other_complex_alias_path", "alias_path", 1),
            ),
        )

    config = AppConfig()
    assert config.alias_config_data == "other_alias_path_2"

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(
            validation_alias=AliasChoices(
                AliasPath("other_complex_alias_path", "Alias_Path", 1),
            ),
        )

    config = AppConfig()
    assert config.alias_config_data == "other_alias_path_2"


# @pytest.mark.current
@pytest.mark.case_sensitive
def test_double_field_case_sensitive(cwd_to_tmp, path_base_dir):
    class AppConfig(ArFiSettings):
        mode: str = "test_mode"

    config = AppConfig()
    assert config.MODE is None
    assert config.mode == "test_mode"

    data = {"mode": "lower_mode"}
    config = AppConfig(**data)
    assert config.MODE == "lower_mode"
    assert config.mode == "lower_mode"

    data = {"MODE": "UPPER_MODE"}
    config = AppConfig(**data)
    assert config.MODE == "UPPER_MODE"
    assert config.mode == "test_mode"

    data = {"MODE": "UPPER_MODE", "mode": "lower_mode"}
    config = AppConfig(**data)
    assert config.MODE == "UPPER_MODE"
    assert config.mode == "lower_mode"


# @pytest.mark.current
@pytest.mark.case_sensitive
def test_double_field_case_sensitive_discriminator(cwd_to_tmp, path_base_dir):
    class MySQL(ArFiSettings):
        DB_TYPE: Literal["mysql"] = "mysql"
        db_type: str = "mysql_default"
        Db_Type: str = "Db_Type_mysql"

    class Postgres(ArFiSettings):
        DB_TYPE: Literal["postgres"] = "postgres"
        db_type: str = "postgres_default"
        Db_Type: str = "Db_Type_postgres"

    class AppConfig(ArFiSettings):
        db: Postgres | MySQL = Field(MySQL(), discriminator="DB_TYPE")

    data = {"db": {"db_type": "postgres"}}
    config = AppConfig(**data)
    assert config.db.DB_TYPE == "postgres"
    assert config.db.db_type == "postgres"
    assert config.db.Db_Type == "Db_Type_postgres"

    data = {"db": {"Db_Type": "postgres"}}
    config = AppConfig(**data)
    assert config.db.DB_TYPE == "postgres"
    assert config.db.db_type == "postgres_default"
    assert config.db.Db_Type == "postgres"

    data = {"db": {"DB_TYPE": "postgres"}}
    config = AppConfig(**data)
    assert config.db.DB_TYPE == "postgres"
    assert config.db.db_type == "postgres_default"
    assert config.db.Db_Type == "Db_Type_postgres"

    data = {"db": {"DB_TYPE": "postgres", "db_type": "INIT_DTATA"}}
    config = AppConfig(**data)
    assert config.db.DB_TYPE == "postgres"
    assert config.db.db_type == "INIT_DTATA"
    assert config.db.Db_Type == "Db_Type_postgres"

    class AppConfig(ArFiSettings):
        db: Postgres | MySQL = Field(MySQL(), discriminator="DB_TYPE")

        model_config = SettingsConfigDict(case_sensitive=True)

    data = {"db": {"db_type": "postgres"}}
    config = AppConfig(**data)
    assert config.db.DB_TYPE == "mysql"
    assert config.db.db_type == "postgres"
    assert config.db.Db_Type == "Db_Type_mysql"
